# -*- coding: utf-8 -*-

"""
These tests all run against an actual Google Music account.

Destructive modifications are not made, but if things go terrible wrong,
an extra test playlist or song may result.
"""
from __future__ import print_function, division, absolute_import, unicode_literals
from builtins import *  # noqa

from collections import namedtuple
import datetime
from hashlib import md5
import itertools
import os
import re
import types
import warnings

from decorator import decorator
from proboscis.asserts import (
    assert_true, assert_equal, assert_is_not_none,
    assert_raises, assert_not_equal, Check,
)
from proboscis import test, before_class, after_class, SkipTest
import requests
from requests.exceptions import SSLError
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from gmusicapi import Musicmanager, Mobileclient
# from gmusicapi.protocol import mobileclient
from gmusicapi.protocol.shared import authtypes
from gmusicapi.utils.utils import retry, id_or_nid
import gmusicapi.test.utils as test_utils

TEST_PLAYLIST_NAME = 'gmusicapi_test_playlist'
TEST_PLAYLIST_DESCRIPTION = 'gmusicapi test playlist'
TEST_STATION_NAME = 'gmusicapi_test_station'

TEST_STORE_GENRE_ID = 'METAL'

# that dumb little intro track on Conspiracy of One,
# picked since it's only a few seconds long
TEST_STORE_SONG_ID = 'Tf3pxtcrp2tw7i6kdxzueoz7uia'

# used for testing streaming.
# differences between clients are presumably from stream quality.
TEST_STORE_SONG_WC_HASH = 'c3302fe6bd54ce9b310f92da1904f3b9'
TEST_STORE_SONG_MC_HASH = '2c05e96d8dd86caa62c2eb3f728a78a8'

# The Nerdist.
TEST_PODCAST_SERIES_ID = 'Iliyrhelw74vdqrro77kq2vrdhy'

# An episode of Note to Self.
# Picked because it's very short (~4 minutes).
TEST_PODCAST_EPISODE_ID = 'Diksw5cywxflebfs3dbiiabfphu'
TEST_PODCAST_EPISODE_HASH = 'e8ff4efd6a3a6a1017b35e0ef564d840'

# Amorphis
TEST_STORE_ARTIST_ID = 'Apoecs6off3y6k4h5nvqqos4b5e'

# Holographic Universe
TEST_STORE_ALBUM_ID = 'B4cao5ms5jjn36notfgnhjtguwa'

# this is owned by my test account, so it shouldn't disappear
TEST_PLAYLIST_SHARETOKEN = ('AMaBXymHAkflgs5lvFAUyyQLYelqqMZNAB4v7Y_-'
                            'v9vmrctLOeW64GScAScoFHEnrLgOP5DSRpl9FYIH'
                            'b84HRBvyIMsxc7Zlrg==')

TEST_CURATED_STATION_ID = 'L75iymnapfmeiklef5rhaqxqiry'

# this is a little data class for the songs we upload
TestSong = namedtuple('TestSong', 'sid title artist album full_data')


def sids(test_songs):
    """Given [TestSong], return ['sid']."""
    return [s.sid for s in test_songs]


def test_subscription_features():
    return 'GM_A' in os.environ


@decorator
def subscription(f, *args, **kwargs):
    """Declare a test to only be run if subscription testing is enabled."""
    if test_subscription_features():
        return f(*args, **kwargs)
    else:
        raise SkipTest('Subscription testing disabled')


@test(groups=['server-other'])
class SslVerificationTests(object):
    test_url = 'https://wrong.host.badssl.com/'

    @test
    def site_has_invalid_cert(self):
        assert_raises(SSLError, requests.head, self.test_url)

    def request_invalid_site(self, client):
        req_kwargs = {'url': self.test_url,
                      'method': 'HEAD'}
        no_auth = authtypes()

        client.session.send(req_kwargs, no_auth)

    @test
    def clients_verify_by_default(self):
        # Webclient removed since testing is disabled.
        for client_cls in (Mobileclient, Musicmanager):
            assert_raises(SSLError, self.request_invalid_site, client_cls())

    @test
    def disable_client_verify(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=InsecureRequestWarning)
            # Webclient removed since testing is disabled.
            for client_cls in (Mobileclient, Musicmanager):
                self.request_invalid_site(client_cls(verify_ssl=False))  # should not raise SSLError


@test(groups=['server'])
class ClientTests(object):
    # set on the instance in login
    # wc = None  # webclient
    mm = None  # musicmanager
    mc = None  # mobileclient

    # These are set on the instance in eg create_song.

    # both are [TestSong]
    user_songs = None
    store_songs = None

    playlist_ids = None
    plentry_ids = None
    station_ids = None

    podcast_ids = None
    delete_podcast = True  # Set to false if user already subscribed to test podcast.

    @property
    def all_songs(self):
        return (self.user_songs or []) + (self.store_songs or [])

    def mc_get_playlist_songs(self, plid):
        """For convenience, since mc can only get all playlists at once."""
        all_contents = self.mc.get_all_user_playlist_contents()
        found = [p for p in all_contents if p['id'] == plid]

        assert_true(len(found), 1)

        return found[0]['tracks']

    @before_class
    def login(self):
        # self.wc = test_utils.new_test_client(Webclient)
        # assert_true(self.wc.is_authenticated())

        self.mm = test_utils.new_test_client(Musicmanager)
        assert_true(self.mm.is_authenticated())

        self.mc = test_utils.new_test_client(Mobileclient)
        assert_true(self.mc.is_authenticated())

    @after_class(always_run=True)
    def logout(self):
        # if self.wc is None:
        #     raise SkipTest('did not create wc')
        # assert_true(self.wc.logout())

        if self.mm is None:
            raise SkipTest('did not create mm')
        assert_true(self.mm.logout())

        if self.mc is None:
            raise SkipTest('did not create mc')
        assert_true(self.mc.logout())

    # This next section is a bit odd: it orders tests that create
    # required resources.

    # The intuitition: starting from an empty library, you need to create
    #  a song before you can eg add it to a playlist.

    # The dependencies end up with an ordering that might look like:
    #
    # with song
    #   with playlist
    #     with plentry
    #   with station
    #
    #
    # Suggestions to improve any of this are welcome!

    @staticmethod
    @retry
    def assert_songs_state(method, sids, present):
        """
        Assert presence/absence of sids and return a list of
        TestSongs found.

        :param method: eg self.mc.get_all_songs
        :param sids: list of song ids
        :param present: if True verify songs are present; False the opposite
        """

        library = method()

        found = [s for s in library if s['id'] in sids]

        expected_len = len(sids)
        if not present:
            expected_len = 0

        assert_equal(len(found), expected_len)

        return [TestSong(s['id'], s['title'], s['artist'], s['album'], s)
                for s in found]

    @staticmethod
    @retry
    def assert_list_inc_equivalence(method, **kwargs):
        """
        Assert that some listing method returns the same
        contents for incremental=True/False.

        :param method: eg self.mc.get_all_songs, must support `incremental` kwarg
        :param **kwargs: passed to method
        """

        lib_chunk_gen = method(incremental=True, **kwargs)
        assert_true(isinstance(lib_chunk_gen, types.GeneratorType))

        assert_equal([e for chunk in lib_chunk_gen for e in chunk],
                     method(incremental=False, **kwargs))

    @test
    def song_create(self):
        # This can create more than one song: one through uploading, one through
        # adding a store track to the library.

        user_sids = []
        store_sids = []

        fname = test_utils.small_mp3

        uploaded, matched, not_uploaded = self.mm.upload(fname)

        if len(not_uploaded) == 1 and 'ALREADY_EXISTS' in not_uploaded[fname]:
            # delete the song if it exists already because a previous test failed
            self.mc.delete_songs(re.search(r'\(.*\)', not_uploaded[fname]).group().strip('()'))

            # and retry the upload
            uploaded, matched, not_uploaded = self.mm.upload(fname)

        # Otherwise, it should have been uploaded normally.
        assert_equal(not_uploaded, {})
        assert_equal(matched, {})
        assert_equal(list(uploaded.keys()), [fname])

        user_sids.append(uploaded[fname])

        if test_subscription_features():
            store_sids.append(self.mc.add_store_tracks(TEST_STORE_SONG_ID)[0])

        # we test get_all_songs here so that we can assume the existance
        # of the song for future tests (the servers take time to sync an upload)

        self.user_songs = self.assert_songs_state(self.mc.get_all_songs, user_sids, present=True)
        self.store_songs = self.assert_songs_state(self.mc.get_all_songs, store_sids, present=True)

    @test
    def playlist_create(self):
        mc_id = self.mc.create_playlist(TEST_PLAYLIST_NAME, "", public=True)
        # wc_id = self.wc.create_playlist(TEST_PLAYLIST_NAME, "", public=True)

        # like song_create, retry until the playlist appears
        @retry
        def assert_playlist_exists(plids):
            found = [p for p in self.mc.get_all_playlists()
                     if p['id'] in plids]

            assert_equal(len(found), 1)

        assert_playlist_exists([mc_id])
        self.playlist_ids = [mc_id]

    @test(depends_on=[playlist_create, song_create],
          runs_after_groups=['playlist.exists', 'song.exists'])
    def plentry_create(self):

        song_ids = [self.user_songs[0].sid]

        # create 3 entries total
        # 3 songs is the minimum to fully test reordering, and also includes the
        # duplicate song_id case
        double_id = self.user_songs[0].sid
        if test_subscription_features():
            double_id = TEST_STORE_SONG_ID

        song_ids += [double_id] * 2

        plentry_ids = self.mc.add_songs_to_playlist(self.playlist_ids[0], song_ids)

        @retry
        def assert_plentries_exist(plid, plentry_ids):
            songs = self.mc_get_playlist_songs(plid)
            found = [e for e in songs
                     if e['id'] in plentry_ids]

            assert_equal(len(found), len(plentry_ids))

        assert_plentries_exist(self.playlist_ids[0], plentry_ids)
        self.plentry_ids = plentry_ids

    @test(groups=['plentry'], depends_on=[plentry_create],
          runs_after_groups=['plentry.exists'],
          always_run=True)
    def plentry_delete(self):
        if self.plentry_ids is None:
            raise SkipTest('did not store self.plentry_ids')

        res = self.mc.remove_entries_from_playlist(self.plentry_ids)
        assert_equal(res, self.plentry_ids)

        @retry
        def assert_plentries_removed(plid, entry_ids):
            found = [e for e in self.mc_get_playlist_songs(plid)
                     if e['id'] in entry_ids]

            assert_equal(len(found), 0)

        assert_plentries_removed(self.playlist_ids[0], self.plentry_ids)

    @test(groups=['playlist'], depends_on=[playlist_create],
          runs_after=[plentry_delete],
          runs_after_groups=['playlist.exists'],
          always_run=True)
    def playlist_delete(self):
        if self.playlist_ids is None:
            raise SkipTest('did not store self.playlist_ids')

        for plid in self.playlist_ids:
            res = self.mc.delete_playlist(plid)
            assert_equal(res, plid)

        @retry
        def assert_playlist_does_not_exist(plid):
            found = [p for p in self.mc.get_all_playlists()
                     if p['id'] == plid]

            assert_equal(len(found), 0)

        for plid in self.playlist_ids:
            assert_playlist_does_not_exist(plid)

    @test
    @subscription
    def station_create(self):
        station_ids = []
        for prefix, kwargs in (
                ('Store song', {'track_id': TEST_STORE_SONG_ID}),
                ('Store-added song', {'track_id': self.store_songs[0].sid}),
                ('up song', {'track_id': self.user_songs[0].sid}),
                ('artist', {'artist_id': TEST_STORE_ARTIST_ID}),
                ('album', {'album_id': TEST_STORE_ALBUM_ID}),
                ('genre', {'genre_id': TEST_STORE_GENRE_ID}),
                ('playlist', {'playlist_token': TEST_PLAYLIST_SHARETOKEN}),
                ('curated station', {'curated_station_id': TEST_CURATED_STATION_ID})):
            station_ids.append(
                self.mc.create_station(prefix + ' ' + TEST_STATION_NAME, **kwargs))

        @retry
        def assert_station_exists(station_id):
            stations = self.mc.get_all_stations()

            found = [s for s in stations
                     if s['id'] == station_id]

            assert_equal(len(found), 1)

        for station_id in station_ids:
            assert_station_exists(station_id)

        self.station_ids = station_ids

    @test(groups=['station'], depends_on=[station_create, song_create],
          runs_after_groups=['station.exists', 'song.exists'],
          always_run=True)
    def station_delete(self):
        if self.station_ids is None:
            raise SkipTest('did not store self.station_ids')

        res = self.mc.delete_stations(self.station_ids)
        assert_equal(res, self.station_ids)

        @retry
        def assert_station_deleted(station_id):
            stations = self.mc.get_all_stations()

            found = [s for s in stations
                     if s['id'] == station_id]

            assert_equal(len(found), 0)

        for station_id in self.station_ids:
            assert_station_deleted(station_id)

    @test(groups=['song'], depends_on=[song_create],
          runs_after=[plentry_delete, station_delete],
          runs_after_groups=["song.exists"],
          always_run=True)
    def song_delete(self):
        # split deletion between wc and mc
        # mc is only to run if subscription testing not enabled
        with Check() as check:
            for i, testsong in enumerate(self.all_songs):
                if True:
                    res = self.mc.delete_songs(testsong.sid)
                else:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        res = self.wc.delete_songs(testsong.sid)
                check.equal(res, [testsong.sid])

        self.assert_songs_state(self.mc.get_all_songs, sids(self.all_songs), present=False)

    @test
    def podcast_create(self):
        # Check to make sure podcast doesn't already exist to prevent deletion.
        already_exists = [pc for pc in self.mc.get_all_podcast_series()
                          if pc['seriesId'] == TEST_PODCAST_SERIES_ID]

        if already_exists:
            self.delete_podcast = False

        # like song_create, retry until the podcast appears
        @retry
        def assert_podcast_exists(pcids):
            found = [pc for pc in self.mc.get_all_podcast_series()
                     if pc['seriesId'] in pcids]

            assert_equal(len(found), 1)

        pc_id = self.mc.add_podcast_series(TEST_PODCAST_SERIES_ID)

        assert_podcast_exists([pc_id])
        self.podcast_ids = [pc_id]

    @test(groups=['podcast'], depends_on=[podcast_create],
          runs_after_groups=['podcast.exists'],
          always_run=True)
    def podcast_delete(self):
        if self.podcast_ids is None:
            raise SkipTest('did not store self.podcast_ids')

        if not self.delete_podcast:
            raise SkipTest('not deleting already existing podcast')

        for pcid in self.podcast_ids:
            res = self.mc.delete_podcast_series(pcid)
            assert_equal(res, pcid)

        @retry
        def assert_podcast_does_not_exist(pcid):
            found = [pc for pc in self.mc.get_all_podcast_series()
                     if pc['seriesId'] == pcid]

            assert_equal(len(found), 0)

        for pcid in self.podcast_ids:
            assert_podcast_does_not_exist(pcid)

    # These decorators just prevent setting groups and depends_on over and over.
    # They won't work right with additional settings; if that's needed this
    #  pattern should be factored out.

    # TODO it'd be nice to have per-client test groups
    song_test = test(groups=['song', 'song.exists'], depends_on=[song_create])
    playlist_test = test(groups=['playlist', 'playlist.exists'],
                         depends_on=[playlist_create])
    plentry_test = test(groups=['plentry', 'plentry.exists'],
                        depends_on=[plentry_create])
    station_test = test(groups=['station', 'station.exists'], depends_on=[station_create])
    podcast_test = test(groups=['podcast', 'podcast.exists'], depends_on=[podcast_create])

    # Non-wonky tests resume down here.

    # ---------
    #  MM tests
    # ---------

    def mm_get_quota(self):
        # just testing the call is successful
        self.mm.get_quota()

    @song_test
    def mm_list_new_songs(self):
        # mm only includes user-uploaded songs
        self.assert_songs_state(self.mm.get_uploaded_songs, sids(self.user_songs), present=True)
        self.assert_songs_state(self.mm.get_uploaded_songs, sids(self.store_songs), present=False)

    @test
    def mm_list_songs_inc_equal(self):
        self.assert_list_inc_equivalence(self.mm.get_uploaded_songs)

    @song_test
    def mm_download_song(self):

        @retry
        def assert_download(sid):
            filename, audio = self.mm.download_song(sid)

            # TODO could use original filename to verify this
            # but, when manually checking, got modified title occasionally
            assert_true(filename.endswith('.mp3'))
            assert_is_not_none(audio)

        assert_download(self.user_songs[0].sid)

    # ---------
    #  WC tests
    # ---------

    # @test
    # def wc_get_registered_devices(self):
    #     # no logic; just checking schema
    #     self.wc.get_registered_devices()

    # @test
    # def wc_get_shared_playlist_info(self):
    #     expected = {
    #         u'author': u'gmusic api',
    #         u'description': u'description here',
    #         u'title': u'public title here',
    #         u'num_tracks': 2
    #     }

    #     assert_equal(
    #         self.wc.get_shared_playlist_info(TEST_PLAYLIST_SHARETOKEN),
    #         expected
    #     )

    # @test
    # @subscription
    # def wc_get_store_stream_urls(self):
    #     urls = self.wc.get_stream_urls(TEST_STORE_SONG_ID)

    #     assert_true(len(urls) > 1)

    # @test
    # @subscription
    # def wc_stream_store_track_with_header(self):
    #     audio = self.wc.get_stream_audio(TEST_STORE_SONG_ID, use_range_header=True)

    #     assert_equal(md5(audio).hexdigest(), TEST_STORE_SONG_WC_HASH)

    # @test
    # @subscription
    # def wc_stream_store_track_without_header(self):
    #     audio = self.wc.get_stream_audio(TEST_STORE_SONG_ID, use_range_header=False)

    #     assert_equal(md5(audio).hexdigest(), TEST_STORE_SONG_WC_HASH)

    # @song_test
    # def wc_get_download_info(self):
    #     url, download_count = self.wc.get_song_download_info(self.user_songs[0].sid)

    #     assert_is_not_none(url)

    # @song_test
    # def wc_get_uploaded_stream_urls(self):
    #     urls = self.wc.get_stream_urls(self.user_songs[0].sid)

    #     assert_equal(len(urls), 1)

    #     url = urls[0]

    #     assert_is_not_none(url)
    #     assert_equal(url.split(':')[0], 'https')

    # @song_test
    # def wc_upload_album_art(self):
    #     url = self.wc.upload_album_art(self.user_songs[0].sid, test_utils.image_filename)
    #     assert_equal(url[:4], 'http')
    #     # TODO download the track and verify the metadata changed

    # ---------
    #  MC tests
    # ---------

    @test
    def mc_get_registered_devices(self):
        # no logic; just checking schema
        self.mc.get_registered_devices()

    @test
    def mc_get_browse_podcast_hierarchy(self):
        # no logic; just checking schema
        self.mc.get_browse_podcast_hierarchy()

    @test
    def mc_get_browse_podcast_series(self):
        # no logic; just checking schema
        self.mc.get_browse_podcast_series()

    @test
    def mc_get_listen_now_items(self):
        # no logic; just checking schema
        self.mc.get_listen_now_items()

    @test
    def mc_get_listen_now_situations(self):
        # no logic; just checking schema
        self.mc.get_listen_now_situations()

    @test
    def mc_list_stations_inc_equal(self):
        self.assert_list_inc_equivalence(self.mc.get_all_stations)

    @test
    def mc_list_shared_playlist_entries(self):
        entries = self.mc.get_shared_playlist_contents(TEST_PLAYLIST_SHARETOKEN)
        assert_true(len(entries) > 0)

    @test
    def mc_stream_podcast_episode(self):
        raise SkipTest('podcast ids keep changing')
        # uses frozen device_id
        # url = self.mc.get_podcast_episode_stream_url(TEST_PODCAST_EPISODE_ID)
        # audio = self.mc.session._rsession.get(url).content
        # assert_equal(md5(audio).hexdigest(), TEST_PODCAST_EPISODE_HASH)

    @test
    @subscription
    def mc_stream_store_track(self):
        url = self.mc.get_stream_url(TEST_STORE_SONG_ID)  # uses frozen device_id
        audio = self.mc.session._rsession.get(url).content
        assert_equal(md5(audio).hexdigest(), TEST_STORE_SONG_MC_HASH)

    @song_test
    def mc_get_uploaded_track_stream_url(self):
        url = self.mc.get_stream_url(self.user_songs[0].sid)

        assert_is_not_none(url)
        assert_equal(url[:4], 'http')

    @staticmethod
    @retry
    def _assert_song_key_equal_to(method, sid, key, value):
        """
        :param method: eg self.mc.get_all_songs
        :param sid: song id
        :param key: eg 'rating'
        :param value: eg '1'
        """
        songs = method()

        if not isinstance(songs, list):
            # kind of a hack to support get_track_info as well
            songs = [songs]

        found = [s for s in songs if id_or_nid(s) == sid]

        assert_equal(len(found), 1)

        assert_equal(found[0][key], value)
        return found[0]

    # how can I get the rating key to show up for store tracks?
    # it works in Google's clients!

    # @test
    # @subscription
    # def mc_change_store_song_rating(self):
    #     song = self.mc.get_track_info(TEST_STORE_SONG_ID)

    #     # increment by one but keep in rating range
    #     rating = int(song.get('rating', '0')) + 1
    #     rating = str(rating % 6)

    #     self.mc.rate_songs(song, rating)

    #     self._assert_song_key_equal_to(lambda: self.mc.get_track_info(TEST_STORE_SONG_ID),
    #                              id_or_nid(song),
    #                              song['rating'])

    @song_test
    def mc_change_uploaded_song_rating(self):
        song = self._assert_song_key_equal_to(
            self.mc.get_all_songs,
            self.all_songs[0].sid,
            'rating',
            '0')  # initially unrated

        self.mc.rate_songs(song, 1)

        self._assert_song_key_equal_to(self.mc.get_all_songs, song['id'], 'rating', '1')

        self.mc.rate_songs(song, 0)

    @song_test
    @retry
    def mc_get_promoted_songs(self):
        song = self.mc.get_track_info(TEST_STORE_SONG_ID)

        self.mc.rate_songs(song, 5)

        promoted = self.mc.get_promoted_songs()
        assert_true(len(promoted))

        self.mc.rate_songs(song, 0)

    def _test_increment_playcount(self, sid):
        matching = [t for t in self.mc.get_all_songs()
                    if t['id'] == sid]
        assert_equal(len(matching), 1)

        # playCount is an optional field.
        initial_playcount = matching[0].get('playCount', 0)

        self.mc.increment_song_playcount(sid, 2)

        self._assert_song_key_equal_to(
            self.mc.get_all_songs,
            sid,
            'playCount',
            initial_playcount + 2)

    @song_test
    def mc_increment_uploaded_song_playcount(self):
        self._test_increment_playcount(self.all_songs[0].sid)

    # Fails silently. See https://github.com/simon-weber/gmusicapi/issues/349.
    # @song_test
    # @subscription
    # def mc_increment_store_song_playcount(self):
    #     self._test_increment_playcount(self.all_songs[1].sid)

    @song_test
    def mc_change_uploaded_song_title_fails(self):
        # this used to work, but now only ratings can be changed.
        # this test is here so I can tell if this starts working again.
        song = self.assert_songs_state(self.mc.get_all_songs, [self.all_songs[0].sid],
                                       present=True)[0]

        old_title = song.title
        new_title = old_title + '_mod'
        # Mobileclient.change_song_metadata is deprecated, so
        # ignore its deprecation warning.
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.mc.change_song_metadata({'id': song.sid, 'title': new_title})

        self._assert_song_key_equal_to(self.mc.get_all_songs, song.sid, 'title', old_title)

    @song_test
    def mc_list_songs_inc_equal(self):
        self.assert_list_inc_equivalence(self.mc.get_all_songs)

    @song_test
    def mc_list_songs_updated_after(self):
        songs_last_minute = self.mc.get_all_songs(
            updated_after=datetime.datetime.now() - datetime.timedelta(minutes=1))
        assert_not_equal(len(songs_last_minute), 0)

        all_songs = self.mc.get_all_songs()
        assert_not_equal(len(songs_last_minute), len(all_songs))

    @podcast_test
    def mc_list_podcast_series_inc_equal(self):
        self.assert_list_inc_equivalence(self.mc.get_all_podcast_series)

    @playlist_test
    def mc_list_playlists_inc_equal(self):
        self.assert_list_inc_equivalence(self.mc.get_all_playlists)

    @playlist_test
    def mc_edit_playlist_name(self):
        new_name = TEST_PLAYLIST_NAME + '_mod'
        plid = self.mc.edit_playlist(self.playlist_ids[0], new_name=new_name)
        assert_equal(self.playlist_ids[0], plid)

        @retry  # change takes time to propogate
        def assert_name_equal(plid, name):
            playlists = self.mc.get_all_playlists()

            found = [p for p in playlists if p['id'] == plid]

            assert_equal(len(found), 1)
            assert_equal(found[0]['name'], name)

        assert_name_equal(self.playlist_ids[0], new_name)

        # revert
        self.mc.edit_playlist(self.playlist_ids[0], new_name=TEST_PLAYLIST_NAME)
        assert_name_equal(self.playlist_ids[0], TEST_PLAYLIST_NAME)

    @playlist_test
    def mc_edit_playlist_description(self):
        new_description = TEST_PLAYLIST_DESCRIPTION + '_mod'
        plid = self.mc.edit_playlist(self.playlist_ids[0], new_description=new_description)
        assert_equal(self.playlist_ids[0], plid)

        @retry  # change takes time to propogate
        def assert_description_equal(plid, description):
            playlists = self.mc.get_all_playlists()

            found = [p for p in playlists if p['id'] == plid]

            assert_equal(len(found), 1)
            assert_equal(found[0]['description'], description)

        assert_description_equal(self.playlist_ids[0], new_description)

        # revert
        self.mc.edit_playlist(self.playlist_ids[0], new_description=TEST_PLAYLIST_DESCRIPTION)
        assert_description_equal(self.playlist_ids[0], TEST_PLAYLIST_DESCRIPTION)

    @playlist_test
    def mc_edit_playlist_public(self):
        new_public = False
        plid = self.mc.edit_playlist(self.playlist_ids[0], public=new_public)
        assert_equal(self.playlist_ids[0], plid)

        @retry  # change takes time to propogate
        def assert_public_equal(plid, public):
            playlists = self.mc.get_all_playlists()

            found = [p for p in playlists if p['id'] == plid]

            assert_equal(len(found), 1)
            assert_equal(found[0]['accessControlled'], public)

        assert_public_equal(self.playlist_ids[0], new_public)

        # revert
        self.mc.edit_playlist(self.playlist_ids[0], public=True)
        assert_public_equal(self.playlist_ids[0], True)

    @playlist_test
    def mc_list_playlists_updated_after(self):
        pls_last_minute = self.mc.get_all_playlists(
            updated_after=datetime.datetime.now() - datetime.timedelta(minutes=1))
        assert_not_equal(len(pls_last_minute), 0)
        print(pls_last_minute)

        all_pls = self.mc.get_all_playlists()
        assert_not_equal(len(pls_last_minute), len(all_pls))

    @retry(tries=3)
    def _mc_assert_ple_position(self, entry, pos):
        """
        :param entry: entry dict
        :pos: 0-based position to assert
        """
        pl = self.mc_get_playlist_songs(entry['playlistId'])

        indices = [i for (i, e) in enumerate(pl)
                   if e['id'] == entry['id']]

        assert_equal(len(indices), 1)

        assert_equal(indices[0], pos)

    @retry
    def _mc_test_ple_reodering(self, from_pos, to_pos):
        if from_pos == to_pos:
            raise ValueError('Will not test no-op reordering.')

        pl = self.mc_get_playlist_songs(self.playlist_ids[0])

        from_e = pl[from_pos]

        e_before_new_pos, e_after_new_pos = None, None

        if from_pos < to_pos:
            adj = 0
        else:
            adj = -1

        if to_pos - 1 >= 0:
            e_before_new_pos = pl[to_pos + adj]

        if to_pos + 1 < len(self.plentry_ids):
            e_after_new_pos = pl[to_pos + adj + 1]

        self.mc.reorder_playlist_entry(from_e,
                                       to_follow_entry=e_before_new_pos,
                                       to_precede_entry=e_after_new_pos)
        self._mc_assert_ple_position(from_e, to_pos)

        if e_before_new_pos:
            self._mc_assert_ple_position(e_before_new_pos, to_pos - 1)

        if e_after_new_pos:
            self._mc_assert_ple_position(e_after_new_pos, to_pos + 1)

    @plentry_test
    def mc_reorder_ple_forwards(self):
        for from_pos, to_pos in [pair for pair in
                                 itertools.product(range(len(self.plentry_ids)), repeat=2)
                                 if pair[0] < pair[1]]:
            self._mc_test_ple_reodering(from_pos, to_pos)

    @plentry_test
    def mc_reorder_ple_backwards(self):
        playlist_len = len(self.plentry_ids)
        for from_pos, to_pos in [pair for pair in
                                 itertools.product(range(playlist_len), repeat=2)
                                 if pair[0] > pair[1]]:
            self._mc_test_ple_reodering(from_pos, to_pos)

    # This fails, unfortunately, which means n reorderings mean n
    # separate calls in the general case.
    # @plentry_test
    # def mc_reorder_ples_forwards(self):
    #    pl = self.mc_get_playlist_songs(self.playlist_ids[0])
    #    # rot2, eg 0123 -> 2301
    #    pl.append(pl.pop(0))
    #    pl.append(pl.pop(0))

    #    mutate_call = mobileclient.BatchMutatePlaylistEntries
    #    mutations = [
    #        mutate_call.build_plentry_reorder(
    #            pl[-1], pl[-2]['clientId'], None),
    #        mutate_call.build_plentry_reorder(
    #            pl[-2], pl[-3]['clientId'], pl[-1]['clientId'])
    #    ]

    #    self.mc._make_call(mutate_call, [mutations])
    #    self._mc_assert_ple_position(pl[-1], len(pl) - 1)
    #    self._mc_assert_ple_position(pl[-2], len(pl) - 2)

    @station_test
    @retry  # sometimes this comes back with no data key
    @subscription
    def mc_list_station_tracks(self):
        for station_id in self.station_ids:
            self.mc.get_station_tracks(station_id, num_tracks=1)
            # used to assert that at least 1 track came back, but
            # our dummy uploaded track won't match anything
            self.mc.get_station_tracks(station_id, num_tracks=1,
                                       recently_played_ids=[TEST_STORE_SONG_ID])
            self.mc.get_station_tracks(station_id, num_tracks=1,
                                       recently_played_ids=[self.user_songs[0].sid])

    def mc_list_IFL_station_tracks(self):
        assert_equal(len(self.mc.get_station_tracks('IFL', num_tracks=1)),
                     1)

    @test(groups=['search'])
    def mc_search_store_no_playlists(self):
        res = self.mc.search('morning', max_results=100)

        # TODO Playlist results are not returned consistently.
        res.pop('playlist_hits')

        with Check() as check:
            for type_, hits in res.items():
                if ((not test_subscription_features() and
                     type_ in ('artist_hits', 'song_hits', 'album_hits'))):
                    # These results aren't returned for non-sub accounts.
                    check.true(len(hits) == 0, "%s had %s hits, expected 0" % (type_, len(hits)))
                else:
                    check.true(len(hits) > 0, "%s had %s hits, expected > 0" % (type_, len(hits)))

    @test
    def mc_artist_info(self):
        aid = 'Apoecs6off3y6k4h5nvqqos4b5e'  # amorphis
        optional_keys = set(('albums', 'topTracks', 'related_artists'))

        include_all_res = self.mc.get_artist_info(aid, include_albums=True,
                                                  max_top_tracks=1, max_rel_artist=1)

        no_albums_res = self.mc.get_artist_info(aid, include_albums=False)
        no_rel_res = self.mc.get_artist_info(aid, max_rel_artist=0)
        no_tracks_res = self.mc.get_artist_info(aid, max_top_tracks=0)

        with Check() as check:
            check.true(set(include_all_res.keys()) & optional_keys == optional_keys)

            check.true(set(no_albums_res.keys()) & optional_keys ==
                       optional_keys - {'albums'})
            check.true(set(no_rel_res.keys()) & optional_keys ==
                       optional_keys - {'related_artists'})
            check.true(set(no_tracks_res.keys()) & optional_keys ==
                       optional_keys - {'topTracks'})

    @test
    @retry
    def mc_album_info(self):
        include_tracks = self.mc.get_album_info(TEST_STORE_ALBUM_ID, include_tracks=True)
        no_tracks = self.mc.get_album_info(TEST_STORE_ALBUM_ID, include_tracks=False)

        with Check() as check:
            check.true('tracks' in include_tracks)
            check.true('tracks' not in no_tracks)

            del include_tracks['tracks']
            check.equal(include_tracks, no_tracks)

    @test
    def mc_track_info(self):
        self.mc.get_track_info(TEST_STORE_SONG_ID)  # just for the schema

    @test
    def mc_podcast_series_info(self):
        optional_keys = {'episodes'}

        include_episodes = self.mc.get_podcast_series_info(TEST_PODCAST_SERIES_ID, max_episodes=1)
        no_episodes = self.mc.get_podcast_series_info(TEST_PODCAST_SERIES_ID, max_episodes=0)

        with Check() as check:
            check.true(set(include_episodes.keys()) & optional_keys == optional_keys)

            check.true(set(no_episodes.keys()) & optional_keys ==
                       optional_keys - {'episodes'})

    @test(groups=['genres'])
    def mc_all_genres(self):
        expected_genres = {u'COMEDY_SPOKEN_WORD_OTHER', u'COUNTRY', u'HOLIDAY', u'R_B_SOUL',
                           u'FOLK', u'LATIN', u'CHRISTIAN_GOSPEL', u'ALTERNATIVE_INDIE', u'POP',
                           u'ROCK', u'WORLD', u'VOCAL_EASY_LISTENING', u'HIP_HOP_RAP', u'JAZZ',
                           u'METAL', u'REGGAE_SKA', u'SOUNDTRACKS_CAST_ALBUMS', u'DANCE_ELECTRONIC',
                           u'CLASSICAL', u'NEW_AGE', u'BLUES', u'CHILDREN_MUSIC'}
        res = self.mc.get_genres()
        assert_equal(set([e['id'] for e in res]), expected_genres)

    @test(groups=['genres'])
    def mc_specific_genre(self):
        expected_genres = {u'PROGRESSIVE_METAL', u'CLASSIC_METAL', u'HAIR_METAL', u'INDUSTRIAL',
                           u'ALT_METAL', u'THRASH', u'METALCORE', u'BLACK_DEATH_METAL',
                           u'DOOM_METAL'}
        res = self.mc.get_genres('METAL')
        assert_equal(set([e['id'] for e in res]), expected_genres)

    @test(groups=['genres'])
    def mc_leaf_parent_genre(self):
        assert_equal(self.mc.get_genres('AFRICA'), [])

    @test(groups=['genres'])
    def mc_invalid_parent_genre(self):
        assert_equal(self.mc.get_genres('bogus genre'), [])



# -*- coding: utf-8 -*-

"""Tools for manipulating client-received Google Music data."""
from __future__ import print_function, division, absolute_import, unicode_literals
from builtins import *  # noqa

import operator
import re
import collections

from collections import Counter
from functools import reduce


def get_id_pairs(track_list):
    """Create a list of (sid, eid) tuples from a list of tracks.
    Tracks without an eid will have an eid of None."""

    return [(t["id"], t.get("playlistEntryId")) for t in track_list]


def find_playlist_changes(orig_tracks, modified_tracks):
    """Finds the changes between two playlists.

    Returns a tuple of (deletions, additions, staying).
    Deletions and additions are both Counters of (sid, eid) tuples;
    staying is a set of (sid, eid) tuples.

    :param old: the original playlist.
    :param modified: the modified playlist."""

    s_pairs = get_id_pairs(orig_tracks)

    # Three cases for desired pairs:
    # 1: (sid, eid from this playlist): either no action or add
    #    (if someone adds a dupe from the same playlist)
    # 2: (sid, eid not from this playlist): add
    # 3: (sid, None): add
    d_pairs = get_id_pairs(modified_tracks)

    # Counters are multisets.
    s_count = Counter(s_pairs)
    d_count = Counter(d_pairs)

    to_del = s_count - d_count
    to_add = d_count - s_count
    to_keep = set(s_count & d_count)  # guaranteed to be counts of 1

    return (to_del, to_add, to_keep)


def filter_song_md(song, md_list=['id'], no_singletons=True):
    """Returns a list of desired metadata from a song.
    Does not modify the given song.

    :param song: Dictionary representing a GM song.
    :param md_list: (optional) the ordered list of metadata to select.
    :param no_singletons: (optional) if md_list is of length 1, return the data,
      not a singleton list.
    """

    filtered = [song[md_type] for md_type in md_list]

    if len(md_list) == 1 and no_singletons:
        return filtered[0]
    else:
        return filtered


def build_song_rep(song, md_list=['title', 'artist', 'album'], divider=" - "):
    """Returns a string of the requested metadata types.
    The order of md_list determines order in the string.

    :param song: Dictionary representing a GM song.
    :param md_list: (optional) list of valid GM metadata types.
    :param divider: (optional) string to join the metadata.
    """

    filtered = filter_song_md(song, md_list, no_singletons=False)

    return divider.join(filtered)


def reorder_to(l, order):
    """Returns a list, reordered to a specific ordering.

    :param l: the list to reorder. It is not modified.
    :param order: a list containing the new ordering,
                  eg [2,1,0] to reverse a list of length 3
    """

    # Zip on ordering, sort by it, then remove ordering.
    return [el[1] for el in sorted(zip(order, l), key=lambda el: el[0])]


def build_queries_from(f, regex, cap_types, cap_pr, encoding='ascii'):
    """Returns a list of queries from the given file.
    Queries have the form [(<query>, <metadata type>), ...]

    :param f: opened file, ready to read.
    :param regex: a compiled regex to capture query info from file lines.
    :param cap_types: the GM metadata types of the regex captures.
    :param cap_pr: the priority of the captures.
    :param encoding: (optional) encoding of the file.
    """

    queries = []

    for line in f:
            matches = regex.match(line)

            if matches:
                # Zip captures to their types and order by priority to build a query.
                query = reorder_to(
                    list(zip(matches.groups(), cap_types)),
                    cap_pr)

                queries.append(query)

    return queries


def build_query_rep(query, divider=" - "):
    """Build a string representation of a query, without metadata types"""

    return divider.join([el[0] for el in query])


# Not mine. From: http://en.wikipedia.org/wiki/Function_composition_(computer_science)
def compose(*funcs, **kfuncs):
    """Compose a group of functions (f(g(h(..)))) into (fogoh...)(...)"""
    return reduce(lambda f, g: lambda *args, **kaargs: f(g(*args, **kaargs)), funcs)


class SongMatcher(object):
    """Matches GM songs to user-provided metadata."""

    def __init__(self, songs, log_metadata=['title', 'artist', 'album']):
        """Prepares songs for matching and determines logging options.

        :param songs: list of GM songs to match against.
        :param log_metadata: list of valid GM metadata types to show in the log.
                             order given will be order outputted.
        """

        # If match times are a problem, could
        # read to an indexed format here.
        self.library = songs

        # Lines of a log of how matching went.
        self.log_lines = []

        self.log_metadata = log_metadata

    def build_log(self):
        """Returns a string built from the current log lines."""

        encoded_lines = [line.encode('utf-8') for line in self.log_lines]
        return "\n".join(encoded_lines)

    def build_song_for_log(self, song):
        """Returns a string built from a song using log options.

        :param song:
        """

        return build_song_rep(song, self.log_metadata)

    class SearchModifier(object):
        """Controls how to query the library.
        Implementations define a comparator, and 2 functions
        (transformers) to modify the query and song data on the fly.

        Sometimes it makes sense to chain implementations.
        In this case, transformers are composed and the most
        outward comparator is used.
        """

        def __init__(self, q_t, s_t, comp):
            # Comparator - defines how to compare query and song data.
            # f(song data, query) -> truthy value
            self.comp = comp

            # Query and song transformers -
            # manipulate query, song before comparison.
            # f(unicode) -> unicode
            self.q_t = q_t

            self.s_t = s_t

    # Some modifiers that are useful in my library:
    # Ignore capitalization:
    ignore_caps = SearchModifier(
        # Change query and song to lowercase,
        # before comparing with ==.
        str.lower,
        str.lower,
        operator.eq
    )

    # Wildcard punctuation (also non ascii chars):
    ignore_punc = SearchModifier(
        # Replace query with a regex, where punc matches any (or no) characters.
        lambda q: re.sub(r"[^a-zA-Z0-9\s]", ".*", q),
        # Don't change the song.
        lambda s: s,
        # The comparator becomes regex matching.
        lambda sd, q: re.search(q, sd)
    )

    implemented_modifiers = (ignore_caps, ignore_punc)

    # The modifiers and order to be used in auto query mode.
    auto_modifiers = implemented_modifiers

    # Tiebreakers are used when there are multiple results from a query.
    @staticmethod
    def manual_tiebreak(query, results):
        """Prompts a user to choose a result from multiple.
        For use with query_library as a tiebreaker.
        Returns a singleton list or None.

        :param query: the original query.
        :param results: list of results.
        """

        print()
        print("Manual tiebreak for query:")
        print(build_query_rep(query).encode('utf-8'))
        print()
        print("Enter the number next to your choice:")
        print()
        print("0: None of these.")

        menu_lines = []
        key = 1

        for song in results:
            menu_lines.append(
                str(key) +
                ": " +
                build_song_rep(song).encode('utf-8'))

            key += 1

        print("\n".join(menu_lines))

        choice = -1

        while not (0 <= choice <= len(results)):
            try:
                choice = int(input("Choice: "))
            except ValueError:
                pass

        return None if choice == 0 else [results[choice - 1]]

    # Tiebreaker which does nothing with results.
    @staticmethod
    def no_tiebreak(query, results):
        return results

    # Exception thrown when a tie is broken.
    class TieBroken(Exception):
        def __init__(self, results):
            self.results = results

    # A named tuple to hold the frozen args when querying recursively.
    QueryState = collections.namedtuple('QueryState', 'orig t_breaker mods auto')

    def query_library(self, query, tie_breaker=no_tiebreak, modifiers=None, auto=False):
        """Queries the library for songs.
        returns a list of matches, or None.
        """

        if not modifiers:
            modifiers = []

        try:
            if not auto:
                return self.query_library_rec(query, self.library,
                                              self.QueryState(query, tie_breaker, modifiers, auto))
            else:
                # Auto mode attempts a search with the current modifiers.
                # If we get 1 result, we return it.
                # If we get no results, we add the next mod from auto_modifers and try again.
                # If we get many results, we branch and try with another modifier.
                # On no results, we tiebreak our old results.
                # Otherwise, we return the branched results.

                current_mods = modifiers[:]
                # Be ready to use any mods from the auto list which we aren't using already.
                future_mods = (m for m in self.auto_modifiers if m not in modifiers)

                while True:  # broken when future_mods runs out

                    # will not break ties in auto mode
                    results = self.query_library_rec(
                        query, self.library,
                        self.QueryState(query, tie_breaker, current_mods, auto))

                    if not results:
                        try:
                            current_mods.append(next(future_mods))
                        except StopIteration:
                            return results

                    elif len(results) == 1:
                        return results

                    else:
                        # Received many results from our current search.
                        # Branch; try more modifers to try and improve.
                        # If results, use them; otherwise tiebreak ours.
                        try:
                            current_mods.append(next(future_mods))
                        except StopIteration:
                            raise self.TieBroken(tie_breaker.__func__(query, results))

                        next_results = self.query_library(query, tie_breaker, current_mods, auto)

                        if not next_results:
                            raise self.TieBroken(tie_breaker.__func__(query, results))
                        else:
                            return next_results
        except self.TieBroken as tie:
            return tie.results

    def query_library_rec(self, query, library, state):
        """Returns a list of matches, or None.
        Recursive querying routine for query_library.
        """

        if len(query) == 0:
            return None

        # Composing applies right to left; currently mods are left to right.
        # Reverse then append the default modifier for proper compose order.
        mods_to_apply = [sm for sm in reversed(state.mods)]
        mods_to_apply.append(self.SearchModifier(
            lambda q: q,
            lambda sd: sd,
            operator.eq))

        # Create the transformers by composing all of them.
        q_t = compose(*list(map((lambda sm: sm.q_t), mods_to_apply)))
        s_t = compose(*list(map((lambda sm: sm.s_t), mods_to_apply)))

        # Use the most outward comparator.
        comp = mods_to_apply[0].comp

        q, md_type = query[0]

        # No need to repeatedly transform q.
        q_transformed = q_t(q)

        # GM limits libraries to 20k songs; this isn't a big performance hit.
        results = [s for s in library if comp(s_t(s[md_type]), q_transformed)]

        # Check for immediate return conditions.
        if not results:
            return None

        if len(results) == 1:
            return [results[0]]

        # Try to refine results by querying them with the next metadata in the query.
        next_query = query[1:]

        next_results = self.query_library_rec(next_query, results, state)

        if not next_results:
            # Don't break ties in auto mode; it's handled a level up.
            if not state.auto:
                raise self.TieBroken(state.t_breaker(state.orig, results))
            else:
                return results

        # Now we have multiple for both our query and the next.
        # Always prefer the next query to ours.
        return next_results

    def match(self, queries, tie_breaker=manual_tiebreak, auto=True):
        """Runs queries against the library; returns a list of songs.
        Match success is logged.

        :param query: list of (query, metadata type) in order of precedence.
                      eg [('The Car Song', 'title'), ('The Cat Empire', 'artist')]
        :param tie_breaker: (optional) tie breaker to use.
        :param modifiers: (optional) An ordered collection of SearchModifers.
          Applied during the query left to right.
        :param auto: (optional) When True, automagically manage modifiers to find results.
        """

        matches = []

        self.log_lines.append("## Starting match of " + str(len(queries)) + " queries ##")

        for query in queries:
            res = self.query_library(query, tie_breaker, auto=auto)

            if res:
                matches += res

            # Log the results.

            # The alert precedes the information for a quick view of what happened.
            alert = None
            if res is None:
                alert = "!!"
            elif len(res) == 1:
                alert = "=="
            else:
                alert = "??"

            # Each query shows the alert and the query.
            self.log_lines.append(alert + " " + build_query_rep(query))

            # Displayed on the line below the alert (might be useful later).
            extra_info = None

            if res:
                for song in res:
                    self.log_lines.append(
                        (extra_info if extra_info else (' ' * len(alert))) +
                        " " +
                        self.build_song_for_log(song))

            elif extra_info:
                self.log_lines.append(extra_info)

        return matches



# -*- coding: utf-8 -*-

from gmusicapi._version import __version__
from gmusicapi.clients import Webclient, Musicmanager, Mobileclient
from gmusicapi.exceptions import CallFailure

__copyright__ = 'Copyright 2015 Simon Weber'
__license__ = 'BSD 3-Clause'
__title__ = 'gmusicapi'

# appease flake8: the imports are purposeful
(__version__, Webclient, Musicmanager, Mobileclient, CallFailure)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import, unicode_literals
from future.utils import PY3, bind_method
from builtins import *  # noqa

from collections import namedtuple
import functools
from getpass import getpass
import logging
import os
import sys

from proboscis import TestProgram

from gmusicapi.clients import Musicmanager, Mobileclient, OAUTH_FILEPATH
from gmusicapi.protocol.musicmanager import credentials_from_refresh_token
from gmusicapi.test import local_tests, server_tests  # noqa
from gmusicapi.test.utils import NoticeLogging

EnvArg = namedtuple('EnvArg', 'envarg kwarg description')

# these names needed to be compressed to fit everything into the travisci key size.
# there's also:
#    * GM_A: when set (to anything) states that we are testing on a subscription account.
#    * GM_AA_D_ID: a registered device id for use with mc streaming

wc_envargs = (
    EnvArg('GM_U', 'email', 'WC user. If not present, user will be prompted.'),
    EnvArg('GM_P', 'password', 'WC password. If not present, user will be prompted.'),
)

mc_envargs = (
    EnvArg('GM_U', 'email', 'MC user. If not present, user will be prompted.'),
    EnvArg('GM_P', 'password', 'MC password. If not present, user will be prompted.'),
    EnvArg('GM_AA_D_ID', 'android_id', 'A registered device id for use with MC streaming'),
)

mm_envargs = (
    EnvArg('GM_O', 'oauth_credentials', 'MM refresh token. Defaults to MM.login default.'),
    EnvArg('GM_I', 'uploader_id', 'MM uploader id. Defaults to MM.login default.'),
    EnvArg('GM_N', 'uploader_name', 'MM uploader name. Default to MM.login default.'),
)


# Webclient auth retreival removed while testing disabled.
#
# def prompt_for_wc_auth():
#     """Return a valid (user, pass) tuple by continually
#     prompting the user."""
#
#     print("These tests will never delete or modify your music."
#           "\n\n"
#           "If the tests fail, you *might* end up with a test"
#           " song/playlist in your library, though."
#           "\n")
#
#     wclient = Webclient()
#     valid_wc_auth = False
#
#     while not valid_wc_auth:
#         print()
#         email = input("Email: ")
#         passwd = getpass()
#
#         valid_wc_auth = wclient.login(email, passwd)
#
#     return email, passwd


def prompt_for_mc_auth():
    """Return a valid (user, pass, android_id) tuple by continually
    prompting the user."""

    print("These tests will never delete or modify your music."
          "\n\n"
          "If the tests fail, you *might* end up with a test"
          " song/playlist in your library, though."
          "\n")

    mclient = Mobileclient()
    valid_mc_auth = False

    while not valid_mc_auth:
        print()
        email = input("Email: ")
        passwd = getpass()

        try:
            android_id = os.environ['GM_AA_D_ID']
        except KeyError:
            android_id = input("Device ID ('mac' for FROM_MAC_ADDRESS): ")

        if android_id == "mac":
            android_id = Mobileclient.FROM_MAC_ADDRESS

        if not android_id:
            print('a device id must be provided')
            sys.exit(1)

        valid_mc_auth = mclient.login(email, passwd, android_id)

    return email, passwd, android_id


def retrieve_auth():
    """Searches the env for auth, prompting the user if necessary.

    On success, return (mc_kwargs, mm_kwargs). On failure, raise ValueError."""

    def get_kwargs(envargs):
        return dict([(arg.kwarg, os.environ.get(arg.envarg))
                     for arg in envargs])

    mc_kwargs = get_kwargs(mc_envargs)
    mm_kwargs = get_kwargs(mm_envargs)

    if not all([mc_kwargs[arg] for arg in ('email', 'password', 'android_id')]):
        if os.environ.get('TRAVIS'):
            print('on Travis but could not read auth from environ; quitting.')
            sys.exit(1)

        mc_kwargs.update(zip(['email', 'password', 'android_id'], prompt_for_mc_auth()))

    if mm_kwargs['oauth_credentials'] is None:
        # ignoring race
        if not os.path.isfile(OAUTH_FILEPATH):
            raise ValueError("You must have oauth credentials stored at the default"
                             " path by Musicmanager.perform_oauth prior to running.")
        del mm_kwargs['oauth_credentials']  # mm default is not None
    else:
        mm_kwargs['oauth_credentials'] = \
            credentials_from_refresh_token(mm_kwargs['oauth_credentials'])

    return (mc_kwargs, mm_kwargs)


def freeze_method_kwargs(klass, method_name, **kwargs):
    method = getattr(klass, method_name)
    partialfunc = functools.partialmethod if PY3 else functools.partial
    bind_method(klass, method_name, partialfunc(method, **kwargs))


def freeze_login_details(mc_kwargs, mm_kwargs):
    """Set the given kwargs to be the default for client login methods."""
    for cls, kwargs in ((Musicmanager, mm_kwargs),
                        (Mobileclient, mc_kwargs),
                        ):
        freeze_method_kwargs(cls, 'login', **kwargs)


def main():
    """Search env for auth envargs and run tests."""

    if '--group=local' not in sys.argv:
        # hack: assume we're just running the proboscis local group
        freeze_login_details(*retrieve_auth())

    # warnings typically signal a change in protocol,
    # so fail the build if anything >= warning are sent,
    noticer = NoticeLogging()
    noticer.setLevel(logging.WARNING)
    root_logger = logging.getLogger('gmusicapi')
    root_logger.addHandler(noticer)

    # proboscis does not have an exit=False equivalent,
    # so SystemExit must be caught instead (we need
    # to check the log noticer)
    try:
        TestProgram(module=sys.modules[__name__]).run_and_exit()
    except SystemExit as e:
        print()
        if noticer.seen_message:
            print('(failing build due to log warnings)')
            sys.exit(1)

        if e.code is not None:
            sys.exit(e.code)

if __name__ == '__main__':
    main()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: uits.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='uits.proto',
  package='wireless_android_skyjam',
  syntax='proto2',
  serialized_pb=_b('\n\nuits.proto\x12\x17wireless_android_skyjam\"\x83\x01\n\tProductId\x12\x35\n\x04type\x18\x01 \x02(\x0e\x32\'.wireless_android_skyjam.ProductId.Type\x12\x18\n\tcompleted\x18\x02 \x01(\x08:\x05\x66\x61lse\x12\n\n\x02id\x18\x03 \x02(\t\"\x19\n\x04Type\x12\x07\n\x03UPC\x10\x01\x12\x08\n\x04GRID\x10\x02\"\\\n\x07\x41ssetId\x12\x33\n\x04type\x18\x01 \x02(\x0e\x32%.wireless_android_skyjam.AssetId.Type\x12\n\n\x02id\x18\x02 \x02(\t\"\x10\n\x04Type\x12\x08\n\x04ISRC\x10\x01\"/\n\rTransactionId\x12\x12\n\x07version\x18\x01 \x02(\t:\x01\x31\x12\n\n\x02id\x18\x02 \x02(\t\"|\n\x07MediaId\x12\x46\n\x0e\x61lgorithm_type\x18\x01 \x02(\x0e\x32..wireless_android_skyjam.MediaId.AlgorithmType\x12\x0c\n\x04hash\x18\x02 \x02(\t\"\x1b\n\rAlgorithmType\x12\n\n\x06SHA256\x10\x01\"\xa3\x01\n\x07UrlInfo\x12\x33\n\x04type\x18\x01 \x02(\x0e\x32%.wireless_android_skyjam.UrlInfo.Type\x12\x0b\n\x03url\x18\x02 \x02(\t\"V\n\x04Type\x12\x08\n\x04WCOM\x10\x01\x12\x08\n\x04WCOP\x10\x02\x12\x08\n\x04WOAF\x10\x03\x12\x08\n\x04WOAR\x10\x04\x12\x08\n\x04WOAS\x10\x05\x12\x08\n\x04WORS\x10\x06\x12\x08\n\x04WPAY\x10\x07\x12\x08\n\x04WPUB\x10\x08\"\xac\x01\n\x0f\x43opyrightStatus\x12;\n\x04type\x18\x01 \x02(\x0e\x32-.wireless_android_skyjam.CopyrightStatus.Type\x12\x11\n\tcopyright\x18\x02 \x01(\t\"I\n\x04Type\x12\x0f\n\x0bUNSPECIFIED\x10\x01\x12\x15\n\x11\x41LLRIGHTSRESERVED\x10\x02\x12\x0e\n\nPRERELEASE\x10\x03\x12\t\n\x05OTHER\x10\x04\"$\n\x05\x45xtra\x12\x0c\n\x04type\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t\"\xf5\x04\n\x0cUitsMetadata\x12\r\n\x05nonce\x18\x01 \x02(\t\x12\x16\n\x0e\x64istributor_id\x18\x02 \x02(\t\x12\x18\n\x10transaction_date\x18\x03 \x02(\t\x12\x36\n\nproduct_id\x18\x04 \x02(\x0b\x32\".wireless_android_skyjam.ProductId\x12\x32\n\x08\x61sset_id\x18\x05 \x02(\x0b\x32 .wireless_android_skyjam.AssetId\x12>\n\x0etransaction_id\x18\x06 \x02(\x0b\x32&.wireless_android_skyjam.TransactionId\x12\x32\n\x08media_id\x18\x07 \x02(\x0b\x32 .wireless_android_skyjam.MediaId\x12\x32\n\x08url_info\x18\x08 \x01(\x0b\x32 .wireless_android_skyjam.UrlInfo\x12Z\n\x16parental_advisory_type\x18\t \x01(\x0e\x32:.wireless_android_skyjam.UitsMetadata.ParentalAdvisoryType\x12\x42\n\x10\x63opyright_status\x18\n \x01(\x0b\x32(.wireless_android_skyjam.CopyrightStatus\x12-\n\x05\x65xtra\x18\x0b \x03(\x0b\x32\x1e.wireless_android_skyjam.Extra\"A\n\x14ParentalAdvisoryType\x12\x0f\n\x0bUNSPECIFIED\x10\x01\x12\x0c\n\x08\x45XPLICIT\x10\x02\x12\n\n\x06\x45\x44ITED\x10\x03\"\xa5\x02\n\rUitsSignature\x12L\n\x0e\x61lgorithm_type\x18\x01 \x02(\x0e\x32\x34.wireless_android_skyjam.UitsSignature.AlgorithmType\x12Z\n\x15\x63\x61nonicalization_type\x18\x02 \x02(\x0e\x32;.wireless_android_skyjam.UitsSignature.CanonicalizationType\x12\x0e\n\x06key_id\x18\x03 \x02(\t\x12\r\n\x05value\x18\x04 \x02(\t\")\n\rAlgorithmType\x12\x0b\n\x07RSA2048\x10\x01\x12\x0b\n\x07\x44SA2048\x10\x02\" \n\x14\x43\x61nonicalizationType\x12\x08\n\x04NONE\x10\x01\"z\n\x04Uits\x12\x37\n\x08metadata\x18\x01 \x02(\x0b\x32%.wireless_android_skyjam.UitsMetadata\x12\x39\n\tsignature\x18\x02 \x02(\x0b\x32&.wireless_android_skyjam.UitsSignature')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_PRODUCTID_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='wireless_android_skyjam.ProductId.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UPC', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GRID', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=146,
  serialized_end=171,
)
_sym_db.RegisterEnumDescriptor(_PRODUCTID_TYPE)

_ASSETID_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='wireless_android_skyjam.AssetId.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ISRC', index=0, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=249,
  serialized_end=265,
)
_sym_db.RegisterEnumDescriptor(_ASSETID_TYPE)

_MEDIAID_ALGORITHMTYPE = _descriptor.EnumDescriptor(
  name='AlgorithmType',
  full_name='wireless_android_skyjam.MediaId.AlgorithmType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SHA256', index=0, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=413,
  serialized_end=440,
)
_sym_db.RegisterEnumDescriptor(_MEDIAID_ALGORITHMTYPE)

_URLINFO_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='wireless_android_skyjam.UrlInfo.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='WCOM', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WCOP', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WOAF', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WOAR', index=3, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WOAS', index=4, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WORS', index=5, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WPAY', index=6, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WPUB', index=7, number=8,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=520,
  serialized_end=606,
)
_sym_db.RegisterEnumDescriptor(_URLINFO_TYPE)

_COPYRIGHTSTATUS_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='wireless_android_skyjam.CopyrightStatus.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALLRIGHTSRESERVED', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PRERELEASE', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OTHER', index=3, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=708,
  serialized_end=781,
)
_sym_db.RegisterEnumDescriptor(_COPYRIGHTSTATUS_TYPE)

_UITSMETADATA_PARENTALADVISORYTYPE = _descriptor.EnumDescriptor(
  name='ParentalAdvisoryType',
  full_name='wireless_android_skyjam.UitsMetadata.ParentalAdvisoryType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXPLICIT', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EDITED', index=2, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1386,
  serialized_end=1451,
)
_sym_db.RegisterEnumDescriptor(_UITSMETADATA_PARENTALADVISORYTYPE)

_UITSSIGNATURE_ALGORITHMTYPE = _descriptor.EnumDescriptor(
  name='AlgorithmType',
  full_name='wireless_android_skyjam.UitsSignature.AlgorithmType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RSA2048', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DSA2048', index=1, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1672,
  serialized_end=1713,
)
_sym_db.RegisterEnumDescriptor(_UITSSIGNATURE_ALGORITHMTYPE)

_UITSSIGNATURE_CANONICALIZATIONTYPE = _descriptor.EnumDescriptor(
  name='CanonicalizationType',
  full_name='wireless_android_skyjam.UitsSignature.CanonicalizationType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NONE', index=0, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1715,
  serialized_end=1747,
)
_sym_db.RegisterEnumDescriptor(_UITSSIGNATURE_CANONICALIZATIONTYPE)


_PRODUCTID = _descriptor.Descriptor(
  name='ProductId',
  full_name='wireless_android_skyjam.ProductId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='wireless_android_skyjam.ProductId.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='completed', full_name='wireless_android_skyjam.ProductId.completed', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='wireless_android_skyjam.ProductId.id', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PRODUCTID_TYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40,
  serialized_end=171,
)


_ASSETID = _descriptor.Descriptor(
  name='AssetId',
  full_name='wireless_android_skyjam.AssetId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='wireless_android_skyjam.AssetId.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='wireless_android_skyjam.AssetId.id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ASSETID_TYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=173,
  serialized_end=265,
)


_TRANSACTIONID = _descriptor.Descriptor(
  name='TransactionId',
  full_name='wireless_android_skyjam.TransactionId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='wireless_android_skyjam.TransactionId.version', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=True, default_value=_b("1").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='wireless_android_skyjam.TransactionId.id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=267,
  serialized_end=314,
)


_MEDIAID = _descriptor.Descriptor(
  name='MediaId',
  full_name='wireless_android_skyjam.MediaId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='algorithm_type', full_name='wireless_android_skyjam.MediaId.algorithm_type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hash', full_name='wireless_android_skyjam.MediaId.hash', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MEDIAID_ALGORITHMTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=316,
  serialized_end=440,
)


_URLINFO = _descriptor.Descriptor(
  name='UrlInfo',
  full_name='wireless_android_skyjam.UrlInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='wireless_android_skyjam.UrlInfo.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='url', full_name='wireless_android_skyjam.UrlInfo.url', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _URLINFO_TYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=443,
  serialized_end=606,
)


_COPYRIGHTSTATUS = _descriptor.Descriptor(
  name='CopyrightStatus',
  full_name='wireless_android_skyjam.CopyrightStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='wireless_android_skyjam.CopyrightStatus.type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='copyright', full_name='wireless_android_skyjam.CopyrightStatus.copyright', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _COPYRIGHTSTATUS_TYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=609,
  serialized_end=781,
)


_EXTRA = _descriptor.Descriptor(
  name='Extra',
  full_name='wireless_android_skyjam.Extra',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='wireless_android_skyjam.Extra.type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='wireless_android_skyjam.Extra.value', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=783,
  serialized_end=819,
)


_UITSMETADATA = _descriptor.Descriptor(
  name='UitsMetadata',
  full_name='wireless_android_skyjam.UitsMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nonce', full_name='wireless_android_skyjam.UitsMetadata.nonce', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='distributor_id', full_name='wireless_android_skyjam.UitsMetadata.distributor_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='transaction_date', full_name='wireless_android_skyjam.UitsMetadata.transaction_date', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='product_id', full_name='wireless_android_skyjam.UitsMetadata.product_id', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='asset_id', full_name='wireless_android_skyjam.UitsMetadata.asset_id', index=4,
      number=5, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='wireless_android_skyjam.UitsMetadata.transaction_id', index=5,
      number=6, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='media_id', full_name='wireless_android_skyjam.UitsMetadata.media_id', index=6,
      number=7, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='url_info', full_name='wireless_android_skyjam.UitsMetadata.url_info', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='parental_advisory_type', full_name='wireless_android_skyjam.UitsMetadata.parental_advisory_type', index=8,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='copyright_status', full_name='wireless_android_skyjam.UitsMetadata.copyright_status', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='extra', full_name='wireless_android_skyjam.UitsMetadata.extra', index=10,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _UITSMETADATA_PARENTALADVISORYTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=822,
  serialized_end=1451,
)


_UITSSIGNATURE = _descriptor.Descriptor(
  name='UitsSignature',
  full_name='wireless_android_skyjam.UitsSignature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='algorithm_type', full_name='wireless_android_skyjam.UitsSignature.algorithm_type', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='canonicalization_type', full_name='wireless_android_skyjam.UitsSignature.canonicalization_type', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='key_id', full_name='wireless_android_skyjam.UitsSignature.key_id', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='wireless_android_skyjam.UitsSignature.value', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _UITSSIGNATURE_ALGORITHMTYPE,
    _UITSSIGNATURE_CANONICALIZATIONTYPE,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1454,
  serialized_end=1747,
)


_UITS = _descriptor.Descriptor(
  name='Uits',
  full_name='wireless_android_skyjam.Uits',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='wireless_android_skyjam.Uits.metadata', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='signature', full_name='wireless_android_skyjam.Uits.signature', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1749,
  serialized_end=1871,
)

_PRODUCTID.fields_by_name['type'].enum_type = _PRODUCTID_TYPE
_PRODUCTID_TYPE.containing_type = _PRODUCTID
_ASSETID.fields_by_name['type'].enum_type = _ASSETID_TYPE
_ASSETID_TYPE.containing_type = _ASSETID
_MEDIAID.fields_by_name['algorithm_type'].enum_type = _MEDIAID_ALGORITHMTYPE
_MEDIAID_ALGORITHMTYPE.containing_type = _MEDIAID
_URLINFO.fields_by_name['type'].enum_type = _URLINFO_TYPE
_URLINFO_TYPE.containing_type = _URLINFO
_COPYRIGHTSTATUS.fields_by_name['type'].enum_type = _COPYRIGHTSTATUS_TYPE
_COPYRIGHTSTATUS_TYPE.containing_type = _COPYRIGHTSTATUS
_UITSMETADATA.fields_by_name['product_id'].message_type = _PRODUCTID
_UITSMETADATA.fields_by_name['asset_id'].message_type = _ASSETID
_UITSMETADATA.fields_by_name['transaction_id'].message_type = _TRANSACTIONID
_UITSMETADATA.fields_by_name['media_id'].message_type = _MEDIAID
_UITSMETADATA.fields_by_name['url_info'].message_type = _URLINFO
_UITSMETADATA.fields_by_name['parental_advisory_type'].enum_type = _UITSMETADATA_PARENTALADVISORYTYPE
_UITSMETADATA.fields_by_name['copyright_status'].message_type = _COPYRIGHTSTATUS
_UITSMETADATA.fields_by_name['extra'].message_type = _EXTRA
_UITSMETADATA_PARENTALADVISORYTYPE.containing_type = _UITSMETADATA
_UITSSIGNATURE.fields_by_name['algorithm_type'].enum_type = _UITSSIGNATURE_ALGORITHMTYPE
_UITSSIGNATURE.fields_by_name['canonicalization_type'].enum_type = _UITSSIGNATURE_CANONICALIZATIONTYPE
_UITSSIGNATURE_ALGORITHMTYPE.containing_type = _UITSSIGNATURE
_UITSSIGNATURE_CANONICALIZATIONTYPE.containing_type = _UITSSIGNATURE
_UITS.fields_by_name['metadata'].message_type = _UITSMETADATA
_UITS.fields_by_name['signature'].message_type = _UITSSIGNATURE
DESCRIPTOR.message_types_by_name['ProductId'] = _PRODUCTID
DESCRIPTOR.message_types_by_name['AssetId'] = _ASSETID
DESCRIPTOR.message_types_by_name['TransactionId'] = _TRANSACTIONID
DESCRIPTOR.message_types_by_name['MediaId'] = _MEDIAID
DESCRIPTOR.message_types_by_name['UrlInfo'] = _URLINFO
DESCRIPTOR.message_types_by_name['CopyrightStatus'] = _COPYRIGHTSTATUS
DESCRIPTOR.message_types_by_name['Extra'] = _EXTRA
DESCRIPTOR.message_types_by_name['UitsMetadata'] = _UITSMETADATA
DESCRIPTOR.message_types_by_name['UitsSignature'] = _UITSSIGNATURE
DESCRIPTOR.message_types_by_name['Uits'] = _UITS

ProductId = _reflection.GeneratedProtocolMessageType('ProductId', (_message.Message,), dict(
  DESCRIPTOR = _PRODUCTID,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.ProductId)
  ))
_sym_db.RegisterMessage(ProductId)

AssetId = _reflection.GeneratedProtocolMessageType('AssetId', (_message.Message,), dict(
  DESCRIPTOR = _ASSETID,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.AssetId)
  ))
_sym_db.RegisterMessage(AssetId)

TransactionId = _reflection.GeneratedProtocolMessageType('TransactionId', (_message.Message,), dict(
  DESCRIPTOR = _TRANSACTIONID,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.TransactionId)
  ))
_sym_db.RegisterMessage(TransactionId)

MediaId = _reflection.GeneratedProtocolMessageType('MediaId', (_message.Message,), dict(
  DESCRIPTOR = _MEDIAID,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.MediaId)
  ))
_sym_db.RegisterMessage(MediaId)

UrlInfo = _reflection.GeneratedProtocolMessageType('UrlInfo', (_message.Message,), dict(
  DESCRIPTOR = _URLINFO,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.UrlInfo)
  ))
_sym_db.RegisterMessage(UrlInfo)

CopyrightStatus = _reflection.GeneratedProtocolMessageType('CopyrightStatus', (_message.Message,), dict(
  DESCRIPTOR = _COPYRIGHTSTATUS,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.CopyrightStatus)
  ))
_sym_db.RegisterMessage(CopyrightStatus)

Extra = _reflection.GeneratedProtocolMessageType('Extra', (_message.Message,), dict(
  DESCRIPTOR = _EXTRA,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.Extra)
  ))
_sym_db.RegisterMessage(Extra)

UitsMetadata = _reflection.GeneratedProtocolMessageType('UitsMetadata', (_message.Message,), dict(
  DESCRIPTOR = _UITSMETADATA,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.UitsMetadata)
  ))
_sym_db.RegisterMessage(UitsMetadata)

UitsSignature = _reflection.GeneratedProtocolMessageType('UitsSignature', (_message.Message,), dict(
  DESCRIPTOR = _UITSSIGNATURE,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.UitsSignature)
  ))
_sym_db.RegisterMessage(UitsSignature)

Uits = _reflection.GeneratedProtocolMessageType('Uits', (_message.Message,), dict(
  DESCRIPTOR = _UITS,
  __module__ = 'uits_pb2'
  # @@protoc_insertion_point(class_scope:wireless_android_skyjam.Uits)
  ))
_sym_db.RegisterMessage(Uits)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

"""
Sessions handle the details of authentication and transporting requests.
"""
from __future__ import print_function, division, absolute_import, unicode_literals
from builtins import *  # noqa

from contextlib import closing

import gpsoauth
import httplib2  # included with oauth2client
import mechanicalsoup
import oauth2client
import requests

from gmusicapi.exceptions import (
    AlreadyLoggedIn, NotLoggedIn, CallFailure
)
from gmusicapi.protocol import webclient
from gmusicapi.utils import utils

log = utils.DynamicClientLogger(__name__)


class _Base(object):
    def __init__(self, rsession_setup=None):
        """
        :param rsession_setup: a callable that will be called with
          the backing requests.Session to delegate config to callers.
        """
        self._rsession = requests.Session()

        if rsession_setup is None:
            rsession_setup = lambda x: x  # noqa
        self._rsession_setup = rsession_setup
        self._rsession_setup(self._rsession)

        self.is_authenticated = False

    def _send_with_auth(self, req_kwargs, desired_auth, rsession):
        raise NotImplementedError

    def _send_without_auth(self, req_kwargs, rsession):
        return rsession.request(**req_kwargs)

    def login(self, *args, **kwargs):
        # subclasses extend / use super()
        if self.is_authenticated:
            raise AlreadyLoggedIn

    def logout(self):
        """
        Reset the session to an unauthenticated, default state.
        """
        self._rsession.close()

        self._rsession = requests.Session()
        self._rsession_setup(self._rsession)

        self.is_authenticated = False

    def send(self, req_kwargs, desired_auth, rsession=None):
        """Send a request from a Call using this session's auth.

        :param req_kwargs: kwargs for requests.Session.request
        :param desired_auth: protocol.shared.AuthTypes to attach
        :param rsession: (optional) a requests.Session to use
          (default ``self._rsession`` - this is exposed for test purposes)
        """
        res = None

        if not any(desired_auth):
            if rsession is None:
                # use a throwaway session to ensure it's clean
                with closing(requests.Session()) as new_session:
                    self._rsession_setup(new_session)
                    res = self._send_without_auth(req_kwargs, new_session)
            else:
                res = self._send_without_auth(req_kwargs, rsession)

        else:
            if not self.is_authenticated:
                raise NotLoggedIn

            if rsession is None:
                rsession = self._rsession

            res = self._send_with_auth(req_kwargs, desired_auth, rsession)

        return res


class Webclient(_Base):
    def login(self, email, password, *args, **kwargs):
        """
        Perform serviceloginauth then retrieve webclient cookies.

        :param email:
        :param password:
        """
        super(Webclient, self).login()

        # Google's login form has a bunch of hidden fields I'd rather not deal with manually.
        browser = mechanicalsoup.Browser(soup_config={"features": "html.parser"})

        login_page = browser.get('https://accounts.google.com/ServiceLoginAuth',
                                 params={'service': 'sj',
                                         'continue': 'https://play.google.com/music/listen'})
        form_candidates = login_page.soup.select("form")
        if len(form_candidates) > 1:
            log.error("Google login form dom has changed; there are %s candidate forms:\n%s",
                      len(form_candidates), form_candidates)
            return False

        form = form_candidates[0]
        form.select("#Email")[0]['value'] = email

        response = browser.submit(form, 'https://accounts.google.com/AccountLoginInfo')

        try:
            response.raise_for_status()
        except requests.HTTPError:
            log.exception("submitting login form failed")
            return False

        form_candidates = response.soup.select("form")
        if len(form_candidates) > 1:
            log.error("Google login form dom has changed; there are %s candidate forms:\n%s",
                      len(form_candidates), form_candidates)
            return False

        form = form_candidates[0]
        form.select("#Passwd")[0]['value'] = password

        response = browser.submit(form, 'https://accounts.google.com/ServiceLoginAuth')

        try:
            response.raise_for_status()
        except requests.HTTPError:
            log.exception("submitting login form failed")
            return False

        # We can't use in without .keys(), since international users will see a
        # CookieConflictError.
        if 'SID' not in list(browser.session.cookies.keys()):
            # Invalid auth.
            return False

        self._rsession.cookies.update(browser.session.cookies)
        self.is_authenticated = True

        # Get webclient cookies.
        # They're stored automatically by requests on the webclient session.
        try:
            webclient.Init.perform(self, True)
        except CallFailure:
            log.exception("unable to initialize webclient cookies")
            self.logout()

        return self.is_authenticated

    def _send_with_auth(self, req_kwargs, desired_auth, rsession):
        if desired_auth.xt:
            req_kwargs.setdefault('params', {})

            req_kwargs['params'].update({'u': 0, 'xt': rsession.cookies['xt']})

        return rsession.request(**req_kwargs)


class Mobileclient(_Base):
    def __init__(self, *args, **kwargs):
        super(Mobileclient, self).__init__(*args, **kwargs)
        self._master_token = None
        self._authtoken = None
        self._locale = None
        self._is_subscribed = None

    def login(self, email, password, android_id, *args, **kwargs):
        """
        Get a master token, then use it to get a skyjam OAuth token.

        :param email:
        :param password:
        :param android_id:
        """

        super(Mobileclient, self).login(email, password, android_id, *args, **kwargs)

        res = gpsoauth.perform_master_login(email, password, android_id)
        if 'Token' not in res:
            return False
        self._master_token = res['Token']

        res = gpsoauth.perform_oauth(
            email, self._master_token, android_id,
            service='sj', app='com.google.android.music',
            client_sig='38918a453d07199354f8b19af05ec6562ced5788')
        if 'Auth' not in res:
            return False
        self._authtoken = res['Auth']

        self.is_authenticated = True

        return True

    def _send_with_auth(self, req_kwargs, desired_auth, rsession):
        if desired_auth.oauth:
            # Default to English (United States) if no locale given.
            if not self._locale:
                self._locale = 'en_US'

            # Set locale for all Mobileclient calls.
            req_kwargs.setdefault('params', {})
            req_kwargs['params'].update({'hl': self._locale})

            # As of API v2.5, dv is a required parameter for all calls.
            # The dv value is part of the Android app version number,
            # but setting this to 0 works fine.
            req_kwargs['params'].update({'dv': 0})

            if self._is_subscribed:
                req_kwargs['params'].update({'tier': 'aa'})
            else:
                req_kwargs['params'].update({'tier': 'fr'})

            req_kwargs.setdefault('headers', {})

            # does this expire?
            req_kwargs['headers']['Authorization'] = \
                'GoogleLogin auth=' + self._authtoken

        return rsession.request(**req_kwargs)


class Musicmanager(_Base):
    def __init__(self, *args, **kwargs):
        super(Musicmanager, self).__init__(*args, **kwargs)
        self._oauth_creds = None

    def login(self, oauth_credentials, *args, **kwargs):
        """Store an already-acquired oauth2client.Credentials."""
        super(Musicmanager, self).login()

        try:
            # refresh the token right away to check auth validity
            oauth_credentials.refresh(httplib2.Http())
        except oauth2client.client.Error:
            log.exception("error when refreshing oauth credentials")

        if oauth_credentials.access_token_expired:
            log.info("could not refresh oauth credentials")
            return False

        self._oauth_creds = oauth_credentials
        self.is_authenticated = True

        return self.is_authenticated

    def _send_with_auth(self, req_kwargs, desired_auth, rsession):
        if desired_auth.oauth:
            if self._oauth_creds.access_token_expired:
                self._oauth_creds.refresh(httplib2.Http())

            req_kwargs['headers'] = req_kwargs.get('headers', {})
            req_kwargs['headers']['Authorization'] = \
                'Bearer ' + self._oauth_creds.access_token

        return rsession.request(**req_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Calls made by the mobile client."""
from __future__ import print_function, division, absolute_import, unicode_literals
from six import raise_from
from builtins import *  # noqa

import base64
import calendar
import copy
from datetime import datetime
from hashlib import sha1
import hmac
import time
from uuid import uuid1

import validictory

import json
from gmusicapi.exceptions import ValidationException, CallFailure
from gmusicapi.protocol.shared import Call, authtypes
from gmusicapi.utils import utils

# URL for sj service
sj_url = 'https://mclients.googleapis.com/sj/v2.5/'
sj_stream_url = 'https://mclients.googleapis.com/music/'

# shared schemas
sj_image_color_styles = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'primary': {'type': 'object',
                    'additionalProperties': False,
                    'properties': {
                        'red': {'type': 'integer'},
                        'green': {'type': 'integer'},
                        'blue': {'type': 'integer'}
                    }},
        'scrim': {'type': 'object',
                  'additionalProperties': False,
                  'properties': {
                      'red': {'type': 'integer'},
                      'green': {'type': 'integer'},
                      'blue': {'type': 'integer'}
                  }},
        'accent': {'type': 'object',
                   'additionalProperties': False,
                   'properties': {
                       'red': {'type': 'integer'},
                       'green': {'type': 'integer'},
                       'blue': {'type': 'integer'}
                   }}
    }
}

sj_image = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'url': {'type': 'string'},
        'aspectRatio': {'type': 'string',
                        'required': False},
        'autogen': {'type': 'boolean',
                    'required': False},
        'colorStyles': sj_image_color_styles.copy()
    }
}

sj_image['properties']['colorStyles']['required'] = False

sj_video = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'id': {'type': 'string'},
        'title': {'type': 'string', 'required': False},
        'thumbnails': {'type': 'array',
                       'required': False,
                       'items': {'type': 'object', 'properties': {
                           'url': {'type': 'string'},
                           'width': {'type': 'integer'},
                           'height': {'type': 'integer'},
                       }}},
    }
}

sj_track = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'title': {'type': 'string'},
        'artist': {'type': 'string'},
        'album': {'type': 'string'},
        'albumArtist': {'type': 'string', 'blank': True},
        'trackNumber': {'type': 'integer'},
        'totalTrackCount': {'type': 'integer', 'required': False},
        'durationMillis': {'type': 'string'},
        'albumArtRef': {'type': 'array',
                        'items': {'type': 'object', 'properties': {'url': {'type': 'string'}}},
                        'required': False},
        'artistArtRef': {'type': 'array',
                         'items': {'type': 'object', 'properties': {'url': {'type': 'string'}}},
                         'required': False,
                         },
        'discNumber': {'type': 'integer'},
        'totalDiscCount': {'type': 'integer', 'required': False},
        'estimatedSize': {'type': 'string'},
        'trackType': {'type': 'string', 'required': False},
        'storeId': {'type': 'string', 'required': False},
        'albumId': {'type': 'string'},
        'artistId': {'type': 'array',
                     'items': {'type': 'string', 'blank': True}, 'required': False},
        'nid': {'type': 'string', 'required': False},
        'trackAvailableForPurchase': {'type': 'boolean', 'required': False},
        'albumAvailableForPurchase': {'type': 'boolean', 'required': False},
        'composer': {'type': 'string', 'blank': True},
        'playCount': {'type': 'integer', 'required': False},
        'year': {'type': 'integer', 'required': False},
        'rating': {'type': 'string', 'required': False},
        'genre': {'type': 'string', 'required': False},
        'trackAvailableForSubscription': {'type': 'boolean', 'required': False},
        'contentType': {'type': 'string'},
        # Only available when rating differs from '0'
        # when using :change_song_metadata:, specifying this value will cause all clients to
        # properly update (web/mobile). As value int(round(time.time() * 1000000)) works quite well
        'lastRatingChangeTimestamp': {'type': 'string', 'required': False},
        'primaryVideo': sj_video.copy(),
        'lastModifiedTimestamp': {'type': 'string', 'required': False},
        'explicitType': {'type': 'string', 'required': False},
        'contentType': {'type': 'string', 'required': False},
        'deleted': {'type': 'boolean', 'required': False},
        'creationTimestamp': {'type': 'string', 'required': False},
        'comment': {'type': 'string', 'required': False, 'blank': True},
        'beatsPerMinute': {'type': 'integer', 'required': False},
        'recentTimestamp': {'type': 'string', 'required': False},
        'clientId': {'type': 'string', 'required': False},
        'id': {'type': 'string', 'required': False}
    }
}
sj_track['properties']['primaryVideo']['required'] = False

sj_playlist = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'name': {'type': 'string'},
        'deleted': {'type': 'boolean',
                    'required': False},  # for public
        'type': {'type': 'string',
                 'pattern': r'MAGIC|SHARED|USER_GENERATED',
                 'required': False,
                 },
        'lastModifiedTimestamp': {'type': 'string',
                                  'required': False},  # for public
        'recentTimestamp': {'type': 'string',
                            'required': False},  # for public
        'shareToken': {'type': 'string'},
        'ownerProfilePhotoUrl': {'type': 'string', 'required': False},
        'ownerName': {'type': 'string',
                      'required': False},
        'accessControlled': {'type': 'boolean',
                             'required': False},  # for public
        'shareState': {'type': 'string',
                       'pattern': r'PRIVATE|PUBLIC',
                       'required': False},  # for public
        'creationTimestamp': {'type': 'string',
                              'required': False},  # for public
        'id': {'type': 'string',
               'required': False},  # for public
        'albumArtRef': {'type': 'array',
                        'items': {'type': 'object', 'properties': {'url': {'type': 'string'}}},
                        'required': False,
                        },
        'description': {'type': 'string',
                        'blank': True,
                        'required': False},
        'explicitType': {'type': 'string', 'required': False},
        'contentType': {'type': 'string', 'required': False}
    }
}

sj_plentry = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'id': {'type': 'string'},
        'clientId': {'type': 'string'},
        'playlistId': {'type': 'string'},
        'absolutePosition': {'type': 'string'},
        'trackId': {'type': 'string'},
        'creationTimestamp': {'type': 'string'},
        'lastModifiedTimestamp': {'type': 'string'},
        'deleted': {'type': 'boolean'},
        'source': {'type': 'string'},
        'track': sj_track.copy()
    },
}

sj_plentry['properties']['track']['required'] = False

sj_attribution = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'license_url': {'type': 'string', 'required': False},
        'license_title': {'type': 'string', 'required': False},
        'source_title': {'type': 'string', 'blank': True, 'required': False},
        'source_url': {'type': 'string', 'blank': True, 'required': False},
    }
}

sj_album = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'name': {'type': 'string'},
        'albumArtist': {'type': 'string'},
        'albumArtRef': {'type': 'string', 'required': False},
        'albumId': {'type': 'string'},
        'artist': {'type': 'string', 'blank': True},
        'artistId': {'type': 'array', 'items': {'type': 'string', 'blank': True}},
        'year': {'type': 'integer', 'required': False},
        'tracks': {'type': 'array', 'items': sj_track, 'required': False},
        'description': {'type': 'string', 'required': False},
        'description_attribution': sj_attribution.copy(),
        'explicitType': {'type': 'string', 'required': False},
        'contentType': {'type': 'string', 'required': False}
    }
}
sj_album['properties']['description_attribution']['required'] = False

sj_artist = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'name': {'type': 'string'},
        'artistArtRef': {'type': 'string', 'required': False},
        'artistArtRefs': {'type': 'array',
                          'items': sj_image,
                          'required': False},
        'artistBio': {'type': 'string', 'required': False},
        'artistId': {'type': 'string', 'blank': True, 'required': False},
        'albums': {'type': 'array', 'items': sj_album, 'required': False},
        'topTracks': {'type': 'array', 'items': sj_track, 'required': False},
        'total_albums': {'type': 'integer', 'required': False},
        'artist_bio_attribution': sj_attribution.copy(),
    }
}

sj_artist['properties']['artist_bio_attribution']['required'] = False
sj_artist['properties']['related_artists'] = {
    'type': 'array',
    'items': sj_artist,  # note the recursion
    'required': False
}

sj_genre = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'id': {'type': 'string'},
        'name': {'type': 'string'},
        'children': {
            'type': 'array',
            'required': False,
            'items': {'type': 'string'}
        },
        'parentId': {
            'type': 'string',
            'required': False,
        },
        'images': {
            'type': 'array',
            'required': False,
            'items': {
                'type': 'object',
                'additionalProperties': False,
                'properties': {
                    'url': {'type': 'string'}
                }
            }
        }
    }
}

sj_station_metadata_seed = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        # one of these will be present
        'artist': {
            'type': sj_artist,
            'required': False
        },
        'genre': {
            'type': sj_genre,
            'required': False
        },
    }
}

sj_station_seed = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'seedType': {'type': 'string'},
        # one of these will be present
        'albumId': {'type': 'string', 'required': False},
        'artistId': {'type': 'string', 'required': False},
        'genreId': {'type': 'string', 'required': False},
        'trackId': {'type': 'string', 'required': False},
        'trackLockerId': {'type': 'string', 'required': False},
        'curatedStationId': {'type': 'string', 'required': False},
        'metadataSeed': {'type': sj_station_metadata_seed, 'required': False}
    }
}

sj_station_track = sj_track.copy()
sj_station_track['properties']['wentryid'] = {'type': 'string', 'required': False}

sj_station = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'imageUrl': {'type': 'string', 'required': False},
        'kind': {'type': 'string'},
        'name': {'type': 'string'},
        'deleted': {'type': 'boolean',
                    'required': False},  # for public
        'lastModifiedTimestamp': {'type': 'string',
                                  'required': False},
        'recentTimestamp': {'type': 'string',
                            'required': False},  # for public
        'clientId': {'type': 'string',
                     'required': False},  # for public
        'sessionToken': {'type': 'string',
                         'required': False},  # for free radios
        'skipEventHistory': {'type': 'array'},  # TODO: What's in this array?
        'seed': sj_station_seed,
        'stationSeeds': {'type': 'array',
                         'items': sj_station_seed},
        'id': {'type': 'string',
               'required': False},  # for public
        'description': {'type': 'string', 'required': False},
        'tracks': {'type': 'array', 'required': False, 'items': sj_station_track},
        'imageUrls': {'type': 'array',
                      'required': False,
                      'items': sj_image
                      },
        'compositeArtRefs': {'type': 'array',
                             'required': False,
                             'items': sj_image
                             },
        'contentTypes': {'type': 'array',
                         'required': False,
                         'items': {'type': 'string'}
                         },
        'byline': {'type': 'string', 'required': False},
        'adTargeting': {
            'type': 'object',
            'properties': {
                'keyword': {'type': 'array',
                            'items': {'type': 'string'}
                            },
            },
            'required': False,
        },
    }
}

sj_listen_now_album = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'artist_metajam_id': {'type': 'string'},
        'artist_name': {'type': 'string'},
        'artist_profile_image': {
            'type': 'object',
            'url': {'type': 'string'}
        },
        'description': {
            'type': 'string',
            'blank': True
        },
        'description_attribution': {
            'type': sj_attribution,
            'required': False
        },
        'explicitType': {'type': 'string', 'required': False},
        'id': {
            'type': 'object',
            'properties': {
                'metajamCompactKey': {'type': 'string'},
                'artist': {'type': 'string'},
                'title': {'type': 'string'}
            }
        },
        'title': {'type': 'string'}
    }
}

sj_listen_now_station = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'highlight_color': {
            'type': 'string',
            'required': False
        },
        'id': {
            'type': 'object',
            'seeds': {
                'type': 'array',
                'items': {'type': sj_station_seed}
            }
        },
        'profile_image': {
            'type': 'object',
            'required': False,
            'url': {'type': 'string'}
        },
        'title': {'type': 'string'}
    }
}

sj_listen_now_item = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'kind': {'type': 'string'},
        'compositeArtRefs': {
            'type': 'array',
            'required': False,
            'items': {'type': sj_image}
        },
        'images': {
            'type': 'array',
            'items': {'type': sj_image},
            'required': False,
        },
        'suggestion_reason': {'type': 'string'},
        'suggestion_text': {'type': 'string'},
        'type': {'type': 'string'},
        'album': {
            'type': sj_listen_now_album,
            'required': False
        },
        'radio_station': {
            'type': sj_listen_now_station,
            'required': False
        }
    }
}

sj_podcast_genre = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'id': {'type': 'string'},
        'displayName': {'type': 'string'}
    }
}

sj_podcast_genre['properties']['subgroups'] = {
    'type': 'array',
    'required': False,
    'items': sj_podcast_genre
}

sj_podcast_episode = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'art': {
            'type': 'array',
            'required': False,
            'items': sj_image
        },
        'author': {
            'type': 'string',
            'required': False
        },
        'deleted': {
            'type': 'string',
            'required': False
        },
        'description': {
            'type': 'string',
            'required': False
        },
        'durationMillis': {'type': 'string'},
        'episodeId': {'type': 'string'},
        'explicitType': {'type': 'string'},
        'fileSize': {'type': 'string'},
        'playbackPositionMillis': {
            'type': 'string',
            'required': False
        },
        'publicationTimestampMillis': {
            'type': 'string',
            'required': False
        },
        'seriesId': {'type': 'string'},
        'seriesTitle': {'type': 'string'},
        'title': {'type': 'string'}
    },
}

sj_podcast_series = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'art': {
            'type': 'array',
            'required': False,
            'items': sj_image
        },
        'author': {'type': 'string'},
        'continuationToken': {
            'type': 'string',
            'required': False,
            'blank': True
        },
        'copyright': {
            'type': 'string',
            'required': False
        },
        'description': {
            'type': 'string',
            'required': False
        },
        'episodes': {
            'type': 'array',
            'required': False,
            'items': sj_podcast_episode
        },
        'explicitType': {'type': 'string'},
        'link': {
            'type': 'string',
            'required': False
        },
        'seriesId': {'type': 'string'},
        'title': {'type': 'string'},
        'totalNumEpisodes': {'type': 'integer'},
        'userPreferences': {
            'type': 'object',
            'required': False,
            'properties': {
                'autoDownload': {
                    'type': 'boolean',
                    'required': False
                },
                'notifyOnNewEpisode': {
                    'type': 'boolean',
                    'required': False
                },
                'subscribed': {'type': 'boolean'}
            }
        }
    }
}

sj_situation = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'description': {'type': 'string'},
        'id': {'type': 'string'},
        'imageUrl': {'type': 'string', 'required': False},
        'title': {'type': 'string'},
        'wideImageUrl': {'type': 'string', 'required': False},
        'stations': {'type': 'array',
                     'required': False,
                     'items': sj_station
                     }
    }
}

sj_situation['properties']['situations'] = {
    'type': 'array',
    'required': False,
    'items': sj_situation
}

sj_search_result = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'score': {'type': 'number', 'required': False},
        'type': {'type': 'string'},
        'best_result': {'type': 'boolean', 'required': False},
        'navigational_result': {'type': 'boolean', 'required': False},
        'navigational_confidence': {'type': 'number', 'required': False},
        'artist': sj_artist.copy(),
        'album': sj_album.copy(),
        'track': sj_track.copy(),
        'playlist': sj_playlist.copy(),
        'series': sj_podcast_series.copy(),
        'station': sj_station.copy(),
        'situation': sj_situation.copy(),
        'youtube_video': sj_video.copy()
    }
}

sj_search_result['properties']['artist']['required'] = False
sj_search_result['properties']['album']['required'] = False
sj_search_result['properties']['track']['required'] = False
sj_search_result['properties']['playlist']['required'] = False
sj_search_result['properties']['series']['required'] = False
sj_search_result['properties']['station']['required'] = False
sj_search_result['properties']['situation']['required'] = False
sj_search_result['properties']['youtube_video']['required'] = False


class McCall(Call):
    """Abstract base for mobile client calls."""

    required_auth = authtypes(oauth=True)

    # validictory schema for the response
    _res_schema = utils.NotImplementedField

    @classmethod
    def validate(cls, response, msg):
        """Use validictory and a static schema (stored in cls._res_schema)."""
        try:
            return validictory.validate(msg, cls._res_schema)
        except ValueError as e:
            raise_from(ValidationException(str(e)), e)

    @classmethod
    def check_success(cls, response, msg):
        # TODO not sure if this is still valid for mc
        pass

        # if 'success' in msg and not msg['success']:
        #    raise CallFailure(
        #        "the server reported failure. This is usually"
        #        " caused by bad arguments, but can also happen if requests"
        #        " are made too quickly (eg creating a playlist then"
        #        " modifying it before the server has created it)",
        #        cls.__name__)

    @classmethod
    def parse_response(cls, response):
        return cls._parse_json(response.text)


class McListCall(McCall):
    """Abc for calls that list a resource."""
    # concrete classes provide:
    item_schema = utils.NotImplementedField
    filter_text = utils.NotImplementedField

    static_headers = {'Content-Type': 'application/json'}
    static_params = {'alt': 'json', 'include-tracks': 'true'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'nextPageToken': {'type': 'string', 'required': False},
            'data': {'type': 'object',
                     'items': {'type': 'array', 'items': item_schema},
                     'required': False,
                     },
        },
    }

    @classmethod
    def dynamic_params(cls, updated_after=None, start_token=None, max_results=None):
        """
        :param updated_after: datetime.datetime; defaults to epoch
        """

        if updated_after is None:
            microseconds = -1
        else:
            microseconds = utils.datetime_to_microseconds(updated_after)

        return {'updated-min': microseconds}

    @classmethod
    def dynamic_data(cls, updated_after=None, start_token=None, max_results=None):
        """
        :param updated_after: ignored
        :param start_token: nextPageToken from a previous response
        :param max_results: a positive int; if not provided, server defaults to 1000
        """
        data = {}

        if start_token is not None:
            data['start-token'] = start_token

        if max_results is not None:
            data['max-results'] = str(max_results)

        return json.dumps(data)

    @classmethod
    def parse_response(cls, response):
        # empty results don't include the data key
        # make sure it's always there
        res = cls._parse_json(response.text)
        if 'data' not in res:
            res['data'] = {'items': []}

        return res

    @classmethod
    def filter_response(cls, msg):
        filtered = copy.deepcopy(msg)
        filtered['data']['items'] = ["<%s %s>" % (len(filtered['data']['items']),
                                                  cls.filter_text)]
        return filtered


class McBatchMutateCall(McCall):
    """Abc for batch mutation calls."""

    static_headers = {'Content-Type': 'application/json'}
    static_params = {'alt': 'json'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'mutate_response': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'id': {'type': 'string', 'required': False},
                        'client_id': {'type': 'string', 'blank': True,
                                      'required': False},
                        'response_code': {'type': 'string'},
                    },
                },
            }
        },
    }

    @staticmethod
    def dynamic_data(mutations):
        """
        :param mutations: list of mutation dictionaries
        """

        return json.dumps({'mutations': mutations})

    @classmethod
    def check_success(cls, response, msg):
        if ('error' in msg or
            not all([d.get('response_code', None) in ('OK', 'CONFLICT')
                     for d in msg.get('mutate_response', [])])):
            raise CallFailure('The server reported failure while'
                              ' changing the requested resource.'
                              " If this wasn't caused by invalid arguments"
                              ' or server flakiness,'
                              ' please open an issue.',
                              cls.__name__)


class McStreamCall(McCall):
    # this call will redirect to the mp3
    static_allow_redirects = False

    _s1 = bytes(base64.b64decode('VzeC4H4h+T2f0VI180nVX8x+Mb5HiTtGnKgH52Otj8ZCGDz9jRW'
                                 'yHb6QXK0JskSiOgzQfwTY5xgLLSdUSreaLVMsVVWfxfa8Rw=='))
    _s2 = bytes(base64.b64decode('ZAPnhUkYwQ6y5DdQxWThbvhJHN8msQ1rqJw0ggKdufQjelrKuiG'
                                 'GJI30aswkgCWTDyHkTGK9ynlqTkJ5L4CiGGUabGeo8M6JTQ=='))

    # bitwise and of _s1 and _s2 ascii, converted to string
    _key = ''.join([chr(c1 ^ c2) for (c1, c2) in zip(_s1, _s2)]).encode("ascii")

    @classmethod
    def get_signature(cls, item_id, salt=None):
        """Return a (sig, salt) pair for url signing."""

        if salt is None:
            salt = str(int(time.time() * 1000))

        mac = hmac.new(cls._key, item_id.encode("utf-8"), sha1)
        mac.update(salt.encode("utf-8"))
        sig = base64.urlsafe_b64encode(mac.digest())[:-1]

        return sig, salt

    @staticmethod
    def dynamic_headers(item_id, device_id, quality):
        return {'X-Device-ID': device_id}

    @classmethod
    def dynamic_params(cls, item_id, device_id, quality):
        sig, salt = cls.get_signature(item_id)

        params = {'opt': quality,
                  'net': 'mob',
                  'pt': 'e',
                  'slt': salt,
                  'sig': sig,
                  }
        if item_id.startswith(('T', 'D')):
            # Store track or podcast episode.
            params['mjck'] = item_id
        else:
            # Library track.
            params['songid'] = item_id

        return params

    @staticmethod
    def parse_response(response):
        return response.headers['location']  # ie where we were redirected

    @classmethod
    def validate(cls, response, msg):
        pass


class Config(McCall):
    static_method = 'GET'
    static_url = sj_url + 'config'

    item_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'key': {'type': 'string'},
            'value': {'type': 'string'}
        }
    }

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'data': {'type': 'object',
                     'entries': {'type': 'array', 'items': item_schema},
                     }
        }
    }


class Search(McCall):
    """Search for All Access tracks."""
    static_method = 'GET'
    static_url = sj_url + 'query'

    # The result types returned are requested in the `ct` parameter.
    # Free account search is restricted so may not contain hits for all result types.
    # 1: Song, 2: Artist, 3: Album, 4: Playlist, 6: Station, 7: Situation, 8: Video
    # 9: Podcast Series
    static_params = {'ct': '1,2,3,4,6,7,8,9'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'clusterOrder': {'type': 'array',
                             'items': {'type': 'string'},
                             'required': False},
            'entries': {'type': 'array',
                        'items': sj_search_result,
                        'required': False},
            'suggestedQuery': {'type': 'string', 'required': False}
        },
    }

    @staticmethod
    def dynamic_params(query, max_results):
        return {'q': query, 'max-results': max_results}


class ListTracks(McListCall):
    item_schema = sj_track
    filter_text = 'tracks'

    static_method = 'POST'
    static_url = sj_url + 'trackfeed'


class GetStreamUrl(McStreamCall):
    static_method = 'GET'
    static_url = sj_stream_url + 'mplay'


class GetStationTrackStreamUrl(McStreamCall):
    static_method = 'GET'
    static_url = sj_stream_url + 'wplay'

    @staticmethod
    def dynamic_headers(item_id, wentry_id, session_token, quality):
        return {'X-Device-ID': ''}

    @classmethod
    def dynamic_params(cls, song_id, wentry_id, session_token, quality):
        sig, salt = cls.get_signature(song_id)

        params = {}
        if song_id[0] == 'T':
            # all access
            params['mjck'] = song_id
        else:
            params['songid'] = song_id

        params['sesstok'] = session_token.encode('utf-8')
        params['wentryid'] = wentry_id.encode('utf-8')
        params['tier'] = 'fr'

        params.update(
            {'audio_formats': 'mp3',
             'opt': quality,
             'net': 'mob',
             'pt': 'a',
             'slt': salt,
             'sig': sig,
             })

        return params

    @staticmethod
    def parse_response(response):
        res = json.loads(response.text)
        return res['location']


class ListPlaylists(McListCall):
    item_schema = sj_playlist
    filter_text = 'playlists'

    static_method = 'POST'
    static_url = sj_url + 'playlistfeed'


class ListPlaylistEntries(McListCall):
    item_schema = sj_plentry
    filter_text = 'plentries'

    static_method = 'POST'
    static_url = sj_url + 'plentryfeed'


class ListSharedPlaylistEntries(McListCall):
    shared_plentry = sj_plentry.copy()
    del shared_plentry['properties']['playlistId']
    del shared_plentry['properties']['clientId']

    item_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'shareToken': {'type': 'string'},
            'responseCode': {'type': 'string'},
            'playlistEntry': {
                'type': 'array',
                'items': shared_plentry,
                'required': False,
            }
        }
    }
    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'entries': {'type': 'array',
                        'items': item_schema,
                        },
        },
    }
    filter_text = 'shared plentries'

    static_method = 'POST'
    static_url = sj_url + 'plentries/shared'

    # odd: this request has an additional level of nesting compared to others,
    # and changes the data/entry schema to entries/playlistEntry.
    # Those horrible naming choices make this even harder to understand.

    @classmethod
    def dynamic_params(cls, share_token, updated_after=None, start_token=None, max_results=None):
        return super(ListSharedPlaylistEntries, cls).dynamic_params(
            updated_after, start_token, max_results)

    @classmethod
    def dynamic_data(cls, share_token, updated_after=None, start_token=None, max_results=None):
        """
        :param share_token: from a shared playlist
        :param updated_after: ignored
        :param start_token: nextPageToken from a previous response
        :param max_results: a positive int; if not provided, server defaults to 1000
        """
        data = {}

        data['shareToken'] = share_token

        if start_token is not None:
            data['start-token'] = start_token

        if max_results is not None:
            data['max-results'] = str(max_results)

        return json.dumps({'entries': [data]})

    @classmethod
    def parse_response(cls, response):
        res = cls._parse_json(response.text)
        if 'playlistEntry' not in res['entries'][0]:
            res['entries'][0]['playlistEntry'] = []

        return res

    @classmethod
    def filter_response(cls, msg):
        filtered = copy.deepcopy(msg)
        filtered['entries'][0]['playlistEntry'] = ["<%s %s>" %
                                                   (len(filtered['entries'][0]['playlistEntry']),
                                                    cls.filter_text)]
        return filtered


class BatchMutatePlaylists(McBatchMutateCall):
    static_method = 'POST'
    static_url = sj_url + 'playlistbatch'

    @staticmethod
    def build_playlist_deletes(playlist_ids):
        # TODO can probably pull this up one
        """
        :param playlist_ids:
        """
        return [{'delete': id} for id in playlist_ids]

    @staticmethod
    def build_playlist_updates(pl_updates):
        """
        :param pl_updates: [{'id': '', name': '', 'description': '', 'public': ''}]
        """
        return [{'update': {
            'id': pl_update['id'],
            'name': pl_update['name'],
            'description': pl_update['description'],
            'shareState': pl_update['public']
        }} for pl_update in pl_updates]

    @staticmethod
    def build_playlist_adds(pl_descriptions):
        """
        :param pl_descriptions: [{'name': '', 'description': '','public': ''}]
        """

        return [{'create': {
            'creationTimestamp': '-1',
            'deleted': False,
            'lastModifiedTimestamp': '0',
            'name': pl_desc['name'],
            'description': pl_desc['description'],
            'type': 'USER_GENERATED',
            'shareState': pl_desc['public'],
        }} for pl_desc in pl_descriptions]


class BatchMutatePlaylistEntries(McBatchMutateCall):
    filter_text = 'plentries'
    item_schema = sj_plentry

    static_method = 'POST'
    static_url = sj_url + 'plentriesbatch'

    @staticmethod
    def build_plentry_deletes(entry_ids):
        """
        :param entry_ids:
        """
        return [{'delete': id} for id in entry_ids]

    @staticmethod
    def build_plentry_reorder(plentry, preceding_cid, following_cid):
        """
        :param plentry: plentry that is moving
        :param preceding_cid: clientid of entry that will be before the moved entry
        :param following_cid: "" that will be after the moved entry
        """

        mutation = copy.deepcopy(plentry)
        keys_to_keep = {'clientId', 'creationTimestamp', 'deleted', 'id', 'lastModifiedTimestamp',
                        'playlistId', 'source', 'trackId'}

        for key in list(mutation.keys()):
            if key not in keys_to_keep:
                del mutation[key]

        # horribly misleading key names; these are _clientids_
        # using entryids works sometimes, but with seemingly random results
        if preceding_cid:
            mutation['precedingEntryId'] = preceding_cid

        if following_cid:
            mutation['followingEntryId'] = following_cid

        return {'update': mutation}

    @staticmethod
    def build_plentry_adds(playlist_id, song_ids):
        """
        :param playlist_id:
        :param song_ids:
        """

        mutations = []

        prev_id, cur_id, next_id = None, str(uuid1()), str(uuid1())

        for i, song_id in enumerate(song_ids):
            m_details = {
                'clientId': cur_id,
                'creationTimestamp': '-1',
                'deleted': False,
                'lastModifiedTimestamp': '0',
                'playlistId': playlist_id,
                'source': 1,
                'trackId': song_id,
            }

            if song_id.startswith('T'):
                m_details['source'] = 2  # AA track

            if i > 0:
                m_details['precedingEntryId'] = prev_id
            if i < len(song_ids) - 1:
                m_details['followingEntryId'] = next_id

            mutations.append({'create': m_details})
            prev_id, cur_id, next_id = cur_id, next_id, str(uuid1())

        return mutations


class GetDeviceManagementInfo(McCall):
    """Get registered device information."""

    static_method = 'GET'
    static_url = sj_url + "devicemanagementinfo"
    static_params = {'alt': 'json'}

    _device_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'id': {'type': 'string'},
            'friendlyName': {'type': 'string', 'blank': True},
            'type': {'type': 'string'},
            'lastAccessedTimeMs': {'type': 'integer'},

            # only for mobile devices
            'smartPhone': {'type': 'boolean', 'required': False},
        }
    }

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'data': {'type': 'object',
                     'items': {'type': 'array', 'items': _device_schema},
                     'required': False,
                     },
        },
    }


class DeauthDevice(McCall):
    """Deauthorize a device from devicemanagementinfo."""

    static_method = 'DELETE'
    static_url = sj_url + "devicemanagementinfo"

    @staticmethod
    def dynamic_params(device_id):
        return {'delete-id': device_id}


class ListPromotedTracks(McListCall):
    item_schema = sj_track
    filter_text = 'tracks'

    static_method = 'POST'
    static_url = sj_url + 'ephemeral/top'


class ListListenNowItems(McCall):
    static_method = 'GET'
    static_url = sj_url + "listennow/getlistennowitems"
    static_params = {'alt': 'json'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'listennow_items': {
                'type': 'array',
                'items': {'type': sj_listen_now_item}
            }
        }
    }

    @staticmethod
    def filter_response(msg):
        filtered = copy.deepcopy(msg)
        if 'listennow_items' in filtered:
            filtered['listennow_items'] = \
                    ["<%s listennow_items>" % len(filtered['listennow_items'])]


class ListListenNowSituations(McCall):
    static_method = 'POST'
    static_url = sj_url + 'listennow/situations'
    static_headers = {'Content-Type': 'application/json'}
    static_params = {'alt': 'json'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'distilledContextWrapper': {'type': 'object',
                                        'distilledContextToken': {'type': 'string'},
                                        'required': False},
            'primaryHeader': {'type': 'string'},
            'subHeader': {'type': 'string'},
            'situations': {'type': 'array',
                           'items': sj_situation,
                           },
        },
    }

    @classmethod
    def dynamic_data(cls):
        tz_offset = calendar.timegm(time.localtime()) - calendar.timegm(time.gmtime())

        return json.dumps({
            'requestSignals': {'timeZoneOffsetSecs': tz_offset}
        })

    @staticmethod
    def filter_response(msg):
        filtered = copy.deepcopy(msg)
        if 'situations' in filtered.get('data', {}):
            filtered['data']['situations'] = \
                    ["<%s situations>" % len(filtered['data']['situations'])]

        return filtered


class GetBrowsePodcastHierarchy(McCall):
    static_method = 'GET'
    static_url = sj_url + 'podcast/browsehierarchy'
    static_params = {'alt': 'json'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'groups': {
                'type': 'array',
                'required': False,  # Only on errors
                'items': sj_podcast_genre
            }
        }
    }


class ListBrowsePodcastSeries(McCall):
    static_method = 'GET'
    static_url = sj_url + 'podcast/browse'
    static_params = {'alt': 'json'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'series': {
                'type': 'array',
                'items': {'type': sj_podcast_series}
            }
        }
    }

    @classmethod
    def dynamic_params(cls, id=None):
        return {'id': id}

    @staticmethod
    def filter_response(msg):
        filtered = copy.deepcopy(msg)
        if 'series' in filtered:
            filtered['series'] = \
                    ["<%s podcasts>" % len(filtered['series'])]


class BatchMutatePodcastSeries(McBatchMutateCall):
    static_method = 'POST'
    static_url = sj_url + 'podcastseries/batchmutate'

    @staticmethod
    def build_podcast_updates(updates):
        """
        :param updates:
          [
            {'seriesId': '', 'subscribed': '', 'userPreferences': {
             'notifyOnNewEpisode': '', 'subscrubed': ''}}...
          ]
        """

        return [{'update': update} for update in updates]


# The podcastseries and podcastepisode list calls are strange in that they require a device
# ID and pass updated-min, max-results, and start-token as params.
# The start-token param is required, even if not given, to get a result for more than one
# call in a session.
class ListPodcastSeries(McListCall):
    item_schema = sj_podcast_series
    filter_text = 'podcast series'

    static_method = 'GET'
    static_url = sj_url + 'podcastseries'

    @staticmethod
    def dynamic_headers(device_id, updated_after=None, start_token=None, max_results=None):
        return {'X-Device-ID': device_id}

    @classmethod
    def dynamic_params(cls, device_id=None, updated_after=None, start_token=None, max_results=None):
        params = {}

        if updated_after is None:
            microseconds = -1
        else:
            microseconds = utils.datetime_to_microseconds(updated_after)

        params['updated-min'] = microseconds

        params['start-token'] = start_token

        if max_results is not None:
            params['max-results'] = str(max_results)

        return params

    @classmethod
    def dynamic_data(cls, device_id=None, updated_after=None, start_token=None, max_results=None):
        pass


# The podcastseries and podcastepisode list calls are strange in that they require a device
# ID and pass updated-min, max-results, and start-token as params.
# The start-token param is required, even if not given, to get a result for more than one
# call in a session.
class ListPodcastEpisodes(McListCall):
    item_schema = sj_podcast_episode
    filter_text = 'podcast episodes'

    static_method = 'GET'
    static_url = sj_url + 'podcastepisode'

    @staticmethod
    def dynamic_headers(device_id, updated_after=None, start_token=None, max_results=None):
        return {'X-Device-ID': device_id}

    @classmethod
    def dynamic_params(cls, device_id=None, updated_after=None, start_token=None, max_results=None):
        params = {}

        if updated_after is None:
            microseconds = -1
        else:
            microseconds = utils.datetime_to_microseconds(updated_after)

        params['updated-min'] = microseconds

        params['start-token'] = start_token

        if max_results is not None:
            params['max-results'] = str(max_results)

        return params

    @classmethod
    def dynamic_data(cls, device_id=None, updated_after=None, start_token=None, max_results=None):
        pass


class GetPodcastEpisodeStreamUrl(McStreamCall):
    static_method = 'GET'
    static_url = sj_stream_url + 'fplay'


class ListStations(McListCall):
    item_schema = sj_station
    filter_text = 'stations'

    static_method = 'POST'
    static_url = sj_url + 'radio/station'


class ListStationTracks(McCall):
    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'data': {'type': 'object',
                     'stations': {'type': 'array', 'items': sj_station},
                     'required': False,
                     },
        },
    }

    static_headers = {'Content-Type': 'application/json'}
    static_params = {'alt': 'json', 'include-tracks': 'true'}
    static_method = 'POST'
    static_url = sj_url + 'radio/stationfeed'

    @staticmethod
    def dynamic_data(station_id, num_entries, recently_played):
        """
        :param station_id:
        :param num_entries: maximum number of tracks to return
        :param recently_played: a list of...song ids? never seen an example
        """
        # TODO
        # clearly, this supports more than one at a time,
        # but then that might introduce paging?
        # I'll leave it for someone else

        if station_id == 'IFL':
            return json.dumps({'contentFilter': 1,
                               'stations': [
                                   {
                                       'numEntries': num_entries,
                                       'recentlyPlayed': recently_played,
                                       'seed': {'seedType': 6}
                                   }
                               ]})

        return json.dumps({'contentFilter': 1,
                           'stations': [
                               {
                                   'numEntries': num_entries,
                                   'radioId': station_id,
                                   'recentlyPlayed': recently_played
                               }
                           ]})

    @staticmethod
    def filter_response(msg):
        filtered = copy.deepcopy(msg)
        if 'stations' in filtered.get('data', {}):
            filtered['data']['stations'] = \
                    ["<%s stations>" % len(filtered['data']['stations'])]
        return filtered


class BatchMutateStations(McBatchMutateCall):
    static_method = 'POST'
    static_url = sj_url + 'radio/editstation'

    @staticmethod
    def build_deletes(station_ids):
        """
        :param station_ids:
        """
        return [{'delete': id, 'includeFeed': False, 'numEntries': 0}
                for id in station_ids]

    @staticmethod
    def build_add(name, seed, include_tracks, num_tracks, recent_datetime=None):
        """
        :param name: the title
        :param seed: a dict {'itemId': id, 'seedType': int}
        :param include_tracks: if True, return `num_tracks` tracks in the response
        :param num_tracks:
        :param recent_datetime: purpose unknown. defaults to now.
        """

        if recent_datetime is None:
            recent_datetime = datetime.now()

        recent_timestamp = utils.datetime_to_microseconds(recent_datetime)

        return {
            "createOrGet": {
                "clientId": str(uuid1()),
                "deleted": False,
                "imageType": 1,
                "lastModifiedTimestamp": "-1",
                "name": name,
                "recentTimestamp": str(recent_timestamp),
                "seed": seed,
                "tracks": []
            },
            "includeFeed": include_tracks,
            "numEntries": num_tracks,
            "params":
            {
                "contentFilter": 1
            }
        }


class BatchMutateTracks(McBatchMutateCall):
    static_method = 'POST'
    static_url = sj_url + 'trackbatch'

    # utility functions to build the mutation dicts
    @staticmethod
    def build_track_deletes(track_ids):
        """
        :param track_ids:
        """
        return [{'delete': id} for id in track_ids]

    @staticmethod
    def build_track_add(store_track_info):
        """
        :param store_track_info: sj_track
        """
        track_dict = copy.deepcopy(store_track_info)
        for key in ('kind', 'trackAvailableForPurchase',
                    'albumAvailableForPurchase', 'albumArtRef',
                    'artistId',
                    ):
            if key in track_dict:
                del track_dict[key]

        for key, default in {
            'playCount': 0,
            'rating': '0',
            'genre': '',
            'lastModifiedTimestamp': '0',
            'deleted': False,
            'beatsPerMinute': -1,
            'composer': '',
            'creationTimestamp': '-1',
            'totalDiscCount': 0,
        }.items():
            track_dict.setdefault(key, default)

        # TODO unsure about this
        track_dict['trackType'] = 8

        return {'create': track_dict}


class GetPodcastSeries(McCall):
    static_method = 'GET'
    static_url = sj_url + 'podcast/fetchseries'
    static_headers = {'Content-Type': 'application/json'}
    static_params = {'alt': 'json'}

    _res_schema = sj_podcast_series

    @staticmethod
    def dynamic_params(podcast_series_id, num_episodes):
        return {
            'nid': podcast_series_id,
            'num': num_episodes}


class GetPodcastEpisode(McCall):
    static_method = 'GET'
    static_url = sj_url + 'podcast/fetchepisode'
    static_headers = {'Content-Type': 'application/json'}
    static_params = {'alt': 'json'}

    _res_schema = sj_podcast_episode

    @staticmethod
    def dynamic_params(podcast_episode_id):
        return {'nid': podcast_episode_id}


class GetStoreTrack(McCall):
    # TODO does this accept library ids, too?
    static_method = 'GET'
    static_url = sj_url + 'fetchtrack'
    static_headers = {'Content-Type': 'application/json'}
    static_params = {'alt': 'json'}

    _res_schema = sj_track

    @staticmethod
    def dynamic_params(track_id):
        return {'nid': track_id}


class GetGenres(McCall):
    static_method = 'GET'
    static_url = sj_url + 'explore/genres'
    static_params = {'alt': 'json'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'kind': {'type': 'string'},
            'genres': {
                'type': 'array',
                'items': sj_genre,
                'required': False,  # only on errors
            }
        }
    }

    @staticmethod
    def dynamic_params(parent_genre_id):
        return {'parent-genre': parent_genre_id}


class GetArtist(McCall):
    static_method = 'GET'
    static_url = sj_url + 'fetchartist'
    static_params = {'alt': 'json'}

    _res_schema = sj_artist

    @staticmethod
    def dynamic_params(artist_id, include_albums, num_top_tracks, num_rel_artist):
        """
        :param include_albums: bool
        :param num_top_tracks: int
        :param num_rel_artist: int
        """

        return {'nid': artist_id,
                'include-albums': include_albums,
                'num-top-tracks': num_top_tracks,
                'num-related-artists': num_rel_artist,
                }


class GetAlbum(McCall):
    static_method = 'GET'
    static_url = sj_url + 'fetchalbum'
    static_params = {'alt': 'json'}

    _res_schema = sj_album

    @staticmethod
    def dynamic_params(album_id, tracks):
        include_tracks = bool(tracks) if tracks else None

        return {'nid': album_id, 'include-tracks': include_tracks}


class IncrementPlayCount(McCall):
    static_method = 'POST'
    static_url = sj_url + 'trackstats'
    static_params = {'alt': 'json'}
    static_headers = {'Content-Type': 'application/json'}

    _res_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'responses': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'additionalProperties': False,
                    'properties': {
                        'id': {'type': 'string',
                               'required': False},  # not provided for AA tracks?
                        'response_code': {'type': 'string'},
                    }
                }
            }
        }
    }

    @staticmethod
    def dynamic_data(sid, plays, playtime):
        # TODO this can support multiple tracks at a time

        play_timestamp = utils.datetime_to_microseconds(playtime)
        event = {
            'context_type': 1,
            'event_timestamp_micros': str(play_timestamp),
            'event_type': 2,
            # This can also send a context_id which is the album/artist id
            # the song was found from.
        }

        return json.dumps({'track_stats': [{
            'id': sid,
            'incremental_plays': plays,
            'last_play_time_millis': str(play_timestamp // 1000),
            'type': 2 if sid.startswith('T') else 1,
            'track_events': [event] * plays,
        }]})



# -*- coding: utf-8 -*-

"""Custom exceptions used across the project."""
from __future__ import print_function, division, absolute_import, unicode_literals
from future.utils import python_2_unicode_compatible
from builtins import *  # noqa


@python_2_unicode_compatible
class CallFailure(Exception):
    """Exception raised when a Google Music server responds that a call failed.

    Attributes:
        callname -- name of the protocol.Call that failed
    """
    def __init__(self, message, callname):
        Exception.__init__(self, message)

        self.callname = callname

    def __str__(self):
        return "%s: %s" % (self.callname, Exception.__str__(self))


class ParseException(Exception):
    """Thrown by Call.parse_response on errors."""
    pass


class ValidationException(Exception):
    """Thrown by Transaction.verify_res_schema on errors."""
    pass


class AlreadyLoggedIn(Exception):
    pass


class NotLoggedIn(Exception):
    pass


class NotSubscribed(Exception):
    def __init__(self, *args):
        if len(args) >= 1:
            args = list(args)
            args[0] += " (https://goo.gl/v1wVHT)"
            args = tuple(args)
        else:
            args = ("Subscription required. (https://goo.gl/v1wVHT)",)

        self.args = args


class GmusicapiWarning(UserWarning):
    pass


class InvalidDeviceId(Exception):
    def __init__(self, message, ids):
        if ids:
            message += 'Your valid device IDs are:\n* %s' % '\n* '.join(ids)
        else:
            message += 'It looks like your account does not have any '
            'valid device IDs.'
        super(InvalidDeviceId, self).__init__(message)
        self.valid_device_ids = ids
